#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : __init__.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 29.09.2021


from    .base           import  TimeOfFlight
from    .base           import  GaussianDensity
from    .base           import  CutpointsToF

from    .cutpoints_tof  import  find_cutpoints_tof


__license__ = "GNU v3.0"
__maintainer__ = "Andrei Leonard Nicusan"
__email__ = "a.l.nicusan@bham.ac.uk"
__status__ = "Beta"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


# File   : pixels.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 23.11.2021


import  time
import  textwrap

import  numpy                   as      np

from    konigcell               import  Pixels
from    pept.utilities.traverse import  traverse2d



def get_cutoff(p1, p2):
    '''Return a numpy array containing the minimum and maximum value found
    across the two input arrays.
    Parameters
    ----------
    p1 : (N,) numpy.ndarray
        The first 1D numpy array.
    p2 : (N,) numpy.ndarray
        The second 1D numpy array.
    Returns
    -------
    (2,) numpy.ndarray
        The minimum and maximum value found across `p1` and `p2`.
    Notes
    -----
    The input parameters *must* be numpy arrays, otherwise an error will
    be raised.
    '''

    return np.array([
        min(p1.min(), p2.min()),
        max(p1.max(), p2.max()),
    ])


def from_lines(
    lines,
    number_of_pixels,
    xlim = None,
    ylim = None,
    verbose = True,
):
    '''Create a pixel space and traverse / pixellise a given sample of
    `lines`.

    The `number_of_pixels` in each dimension must be defined. If the
    pixel space boundaries `xlim` or `ylim` are not defined, they
    are inferred as the boundaries of the `lines`.

    Parameters
    ----------
    lines : (M, N>=5) numpy.ndarray
        The lines that will be pixellised, each defined by a timestamp and
        two 2D points, so that the data columns are [time, x1, y1, x2, y2].
        Note that extra columns are ignored.

    number_of_pixels : (2,) list[int]
        The number of pixels in the x- and y-dimensions, respectively.

    xlim : (2,) list[float], optional
        The lower and upper boundaries of the pixellised volume in the
        x-dimension, formatted as [x_min, x_max]. If undefined, it is
        inferred from the boundaries of `lines`.

    ylim : (2,) list[float], optional
        The lower and upper boundaries of the pixellised volume in the
        y-dimension, formatted as [y_min, y_max]. If undefined, it is
        inferred from the boundaries of `lines`.

    Returns
    -------
    pept.Pixels
        A new `Pixels` object with the pixels through which the lines were
        traversed.

    Raises
    ------
    ValueError
        If the input `lines` does not have the shape (M, N>=5). If the
        `number_of_pixels` is not a 1D list with exactly 2 elements, or
        any dimension has fewer than 2 pixels.
    '''

    if verbose:
        start = time.time()

    # Type-checking inputs
    lines = np.asarray(lines, order = "C", dtype = float)

    if lines.ndim != 2 or lines.shape[1] < 5:
        raise ValueError(textwrap.fill((
            "The input `lines` must be a 2D numpy array containing lines "
            "defined by a timestamp and two 2D points, with every row "
            "formatted as [t, x1, y1, x2, y2]. The `lines` must then have "
            f"shape (M, 5). Received array with shape {lines.shape}."
        )))

    number_of_pixels = np.asarray(
        number_of_pixels,
        order = "C",
        dtype = int
    )

    if number_of_pixels.ndim != 1 or len(number_of_pixels) != 2:
        raise ValueError(textwrap.fill((
            "The input `number_of_pixels` must be a list-like "
            "with exactly two values, corresponding to the "
            "number of pixels in the x- and y-dimension. "
            f"Received parameter with shape {number_of_pixels.shape}."
        )))

    if (number_of_pixels < 2).any():
        raise ValueError(textwrap.fill((
            "The input `number_of_pixels` must set at least two "
            "pixels in each dimension (i.e. all elements in "
            "`number_of_elements` must be larger or equal to two). "
            f"Received `{number_of_pixels}`."
        )))

    if xlim is None:
        xlim = get_cutoff(lines[:, 1], lines[:, 3])
    else:
        xlim = np.asarray(xlim, dtype = float)

        if xlim.ndim != 1 or len(xlim) != 2:
            raise ValueError(textwrap.fill((
                "The input `xlim` parameter must be a list with exactly "
                "two values, corresponding to the minimum and maximum "
                "coordinates of the pixel space in the x-dimension. "
                f"Received parameter with shape {xlim.shape}."
            )))

    if ylim is None:
        ylim = get_cutoff(lines[:, 1], lines[:, 3])
    else:
        ylim = np.asarray(ylim, dtype = float)

        if ylim.ndim != 1 or len(ylim) != 2:
            raise ValueError(textwrap.fill((
                "The input `ylim` parameter must be a list with exactly "
                "two values, corresponding to the minimum and maximum "
                "coordinates of the pixel space in the y-dimension. "
                f"Received parameter with shape {ylim.shape}."
            )))

    pixels_array = np.zeros(tuple(number_of_pixels))
    pixels = Pixels(
        pixels_array,
        xlim = xlim,
        ylim = ylim,
    )

    pixels.add_lines(lines, verbose = False)

    if verbose:
        end = time.time()
        print((
            f"Initialised Pixels class in {end - start} s."
        ))

    return pixels


# Dynamically add this as a Pixels method
def add_lines(self, lines, verbose = False):
    '''Pixellise a sample of lines, adding 1 to each pixel traversed, for
    each line in the sample.

    Parameters
    ----------
    lines : (M, N >= 5) numpy.ndarray
        The sample of 2D lines to pixellise. Each line is defined as a
        timestamp followed by two 2D points, such that the data columns are
        `[time, x1, y1, x2, y2, ...]`. Note that there can be extra data
        columns which will be ignored.

    verbose : bool, default False
        Time the pixel traversal and print it to the terminal.

    Raises
    ------
    ValueError
        If `lines` has fewer than 5 columns.
    '''

    lines = np.asarray(lines, order = "C", dtype = float)
    if lines.ndim != 2 or lines.shape[1] < 5:
        raise ValueError(textwrap.fill((
            "The input `lines` must be a 2D array of lines, where each "
            "line (i.e. row) is defined by a timestamp and two 2D points, "
            "so the data columns are [time, x1, y1, x2, y2]. "
            f"Received array of shape {lines.shape}."
        )))

    if verbose:
        start = time.time()

    traverse2d(
        self.pixels,
        lines,
        self.pixel_grids[0],
        self.pixel_grids[1],
    )

    if verbose:
        end = time.time()
        print(f"Traversing {len(lines)} lines took {end - start} s.")


# Add the `from_lines` function as a static method to Pixels
Pixels.from_lines = staticmethod(from_lines)
Pixels.add_lines = add_lines



# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2022 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools


# File              : sampling_extensions.pyx
# License           : GNU v3.0
# Author            : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date              : 14.08.2022


# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: nonecheck=False
# cython: embedsignature=True
# cython: cdivision=True


import numpy as np      # import numpy for Python functions

from libc.stdint cimport int64_t


cpdef samples_indices_adaptive_window_ext(
    const double[:, :] data,          # LoRs in sample
    const double window,              # Time window
    const double overlap,             # Time window overlap
    const Py_ssize_t max_elems,       # Maximum number of LoRs in a sample
):
    '''Compute the minimum distance points (MDPs) from all combinations of
    `num_lines` lines given in an array of lines `sample_lines`.

    ::

        Function signature:
            find_minpoints(
                double[:, :] sample_lines,  # LoRs in sample
                Py_ssize_t num_lines,       # Number of LoRs in combinations
                double max_distance,        # Max distance from MDP to LoRs
                double[:] cutoffs,          # Spatial cutoff for minpoints
                bool append_indices = 0     # Append LoR indices used
            )

    Given a sample of lines, this functions computes the minimum distance
    points (MDPs) for every possible combination of `num_lines` lines. The
    returned numpy array contains all MDPs that satisfy the following:

    1. Are within the `cutoffs`.
    2. Are closer to all the constituent LoRs than `max_distance`.

    Parameters
    ----------
    sample_lines : (M, N) numpy.ndarray
        A 2D array of lines, where each line is defined by two points such that
        every row is formatted as `[t, x1, y1, z1, x2, y2, z2, etc.]`. It
        *must* have at least 2 lines and the combination size `num_lines`
        *must* be smaller or equal to the number of lines. Put differently:
        2 <= num_lines <= len(sample_lines).

    num_lines : int
        The number of lines in each combination of LoRs used to compute the
        MDP. This function considers every combination of `numlines` from the
        input `sample_lines`. It must be smaller or equal to the number of input
        lines `sample_lines`.

    max_distance : float
        The maximum allowed distance between an MDP and its constituent lines.
        If any distance from the MDP to one of its lines is larger than
        `max_distance`, the MDP is thrown away.

    cutoffs : (6,) numpy.ndarray
        An array of spatial cutoff coordinates with *exactly 6 elements* as
        [x_min, x_max, y_min, y_max, z_min, z_max]. If any MDP lies outside
        this region, it is thrown away.

    append_indices : bool
        A boolean specifying whether to include the indices of the lines used
        to compute each MDP. If `False`, the output array will only contain the
        [time, x, y, z] of the MDPs. If `True`, the output array will have
        extra columns [time, x, y, z, line_idx(1), ..., line_idx(n)] where
        n = `num_lines`.

    Returns
    -------
    minpoints : (M, N) numpy.ndarray
        A 2D array of `float`s containing the time and coordinates of the MDPs
        [time, x, y, z]. The time is computed as the average of the constituent
        lines. If `append_indices` is `True`, then `num_lines` indices of the
        constituent lines are appended as extra columns:
        [time, x, y, z, line_idx1, line_idx2, ..].

    Notes
    -----
    There must be at least two lines in `sample_lines` and `num_lines` must be
    greater or equal to the number of lines (i.e. `len(sample_lines)`).
    Put another way: 2 <= num_lines <= len(sample_lines).

    This is a low-level Cython function that does not do any checks on the
    input data - it is meant to be used in other modules / libraries. For a
    normal user, the `pept.tracking.peptml` function `find_minpoints` and
    class `Minpoints` are recommended as higher-level APIs. They do check the
    input data and are easier to use (for example, they automatically compute
    the cutoffs).

    Examples
    --------

    >>> import numpy as np
    >>> from pept.utilities import find_minpoints
    >>>
    >>> lines = np.random.random((500, 7)) * 500
    >>> num_lines = 3
    >>> max_distance = 0.1
    >>> cutoffs = np.array([0, 500, 0, 500, 0, 500], dtype = float)
    >>>
    >>> minpoints = find_minpoints(lines, num_lines, max_distance, cutoffs)

    '''

    if overlap >= window or max_elems < 1:
        raise ValueError("overlap >= window || max_elems < 1")

    # Approximate number of samples - will still need to check / reallocate as needed
    samples_indices_arr = np.zeros(
        (<int>((data[len(data) - 1, 0] - data[0, 0]) / (window - overlap)), 2),
        dtype = np.int64,
    )

    cdef int64_t[:, :] samples_indices = samples_indices_arr

    # Index and time of the start of the current window
    cdef int64_t istart = 0
    cdef double  tstart = data[0, 0]

    # Index and time of the start of the next window, considering overlap
    cdef int64_t inext = 0
    cdef double  tnext = tstart + window - overlap
    cdef int64_t found_next = 0

    # Line index, sample index, number of lines
    cdef int64_t i = 0
    cdef int64_t isample = 0
    cdef int64_t nrows = data.shape[0]

    while i < nrows:
        # print(f"t={data[i, 0]}")
        # print(f"{i=}\n{isample=}\n{istart=}\n{tstart=}\n{inext=}\n{tnext=}\n")

        # If next window is found (considering overlap)
        if data[i, 0] > tnext and found_next == 0:
            inext = i
            found_next = 1

        # If the end of the current window is found - i.e. time window is
        # exceeded or maximum number of lines is reached
        if data[i, 0] > tstart + window or i - istart >= max_elems:

            # If the next window's start wasn't reached yet, just start here
            if found_next == 0:
                inext = i

            # Set the end index of the current sample
            samples_indices[isample, 1] = i
            isample += 1

            # Move indices to next window's start
            istart = inext
            i = istart

            # The starting time of the next sample
            tstart = data[istart, 0]
            tnext = tstart + window - overlap

            # Reallocate samples_indices if more samples are needed
            if isample >= len(samples_indices):
                samples_indices_new = np.zeros(
                    (2 * len(samples_indices), 2),
                    dtype = np.int64,
                )

                samples_indices_new[:len(samples_indices), :] = samples_indices
                samples_indices_arr = samples_indices_new
                samples_indices = samples_indices_arr

            samples_indices[isample, 0] = istart
            found_next = 0

        i += 1

    samples_indices[isample, 1] = i

    # Truncate the samples indices which were not written to
    truncated = np.delete(samples_indices_arr, slice(isample + 1, None, None), 0)
    del(samples_indices_arr)

    return truncated



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : __init__.py
# License:  GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 22.08.2019


'''Separate the intertwined points from pre-tracked tracer locations into
individual trajectories.

Extended Summary
----------------
A typical PEPT workflow would involve transforming LoRs into points using some
tracking algorithm. These points include all tracers moving through the system,
being intertwined (e.g. for two tracers A and B, the `point_data` array might
have two entries for A, followed by three entries for B, then one entry for A,
etc.). The points can be segregated based on distance alone using the
`segregate_trajectories` function; for well-defined trajectories of tracers
that do not collide, this may be enough to retrieve individual trajectories.
However, for tracers that do come into contact, the identity of the least
active one is usually lost; in such cases, the `connect_trajectories` function
can be used to piece back the trajectories of tracers with gaps in their tracks
using some *tracer signature* (e.g. cluster size in PEPT-ML).
'''


from    .trajectory_separation      import  Segregate
from    .trajectory_separation      import  Reconnect


__license__ = "GNU v3.0"
__maintainer__ = "Andrei Leonard Nicusan"
__email__ = "a.l.nicusan@bham.ac.uk"
__status__ = "Beta"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : mixing.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 24.08.2022


import  os
import  textwrap
from    numbers             import  Number


import  numpy               as      np
import  scipy
from    scipy.interpolate   import  interp1d

from    tqdm                import  tqdm
from    joblib              import  Parallel, delayed

from    pept                import  PointData, Voxels
from    pept.base           import  Reducer
from    pept                import  plots

from    plotly.subplots     import  make_subplots
import  plotly.graph_objs   as      go




def enlarge(lim):
    '''Correctly enlarge two limits, e.g. [xmin, xmax]
    '''
    eps = np.finfo(float).resolution
    lim = lim.copy()

    if lim[0] < 0:
        lim[0] = lim[0] * 1.0001 - eps
    else:
        lim[0] = lim[0] * 0.9999 - eps

    if lim[1] < 0:
        lim[1] = lim[1] * 0.9999 + eps
    else:
        lim[1] = lim[1] * 1.0001 + eps

    return lim




def _check_point3d(**kwargs):
    # Extract first argument
    for name, point in kwargs.items():
        break

    if point.ndim != 1 or len(point) != 3:
        raise ValueError(textwrap.fill((
            f"The input point `{name}` should be a vector-like with exactly 3 "
            f"elements. Received `{point}`."
        )))




def _parallel(f, samples, executor, max_workers, desc = "", verbose = True):
    if executor == "sequential":
        if verbose:
            samples = tqdm(samples, desc = desc)

        return [f(s) for s in samples]

    elif executor == "joblib":
        if verbose:
            samples = tqdm(samples, desc = desc)

        # Joblib's `max_workers` behaviour is different than for Executors.
        # The latter simply sets the maximum number of threads available,
        # while the former uses a max_worker=1. Make behaviour consistent.
        if max_workers is None:
            max_workers = os.cpu_count()

        return Parallel(n_jobs = max_workers)(delayed(f)(s) for s in samples)

    else:
        # Otherwise assume `executor` is a `concurrent.futures.Executor`
        # subclass (e.g. ProcessPoolExecutor, MPIPoolExecutor).
        with executor(max_workers = max_workers) as exe:
            futures = [exe.submit(f, s) for s in samples]

            if verbose:
                futures = tqdm(futures, desc = desc)

            return [fut.result() for fut in futures]




class LaceyColors(Reducer):
    '''Compute Lacey-like mixing image, with streamlines passing through plane
    1 being split into Red and Blue tracers, then evaluated into RGB pixels at
    a later point in plane 2.

    Intuitively, red and blue pixels will contain unmixed streamlines, while
    purple pixels will indicate mixing.

    Reducer signature:

    ::

               PointData -> LaceyColors.fit -> (height, width, 3) pept.Voxels
         list[PointData] -> LaceyColors.fit -> (height, width, 3) pept.Voxels
        list[np.ndarray] -> LaceyColors.fit -> (height, width, 3) pept.Voxels

    **Each sample in the input `PointData` is treated as a separate streamline
    / tracer pass. You can group passes using `Segregate + GroupBy("label")`**.

    The first plane where tracers are split into Red and Blue streamlines is
    defined by a point `p1` and direction axis `ax1`. **The point `p1` should
    be the middle of the pipe**.

    The second plane where mixing is evaluated is similarly defined by `p2` and
    `ax2`. **The point `p2` should be the middle of the volume / pipe**.

    If the direction vectors `ax1` and `ax2` are undefined (`None`), the
    tracers are assumed to follow a straight line between `p1` and `p2`.

    The `max_distance` parameter defines the maximum distance allowed between
    a point and a plane to be considered part of it. The `resolution` defines
    the number of pixels in the height and width of the resulting image.

    *New in pept-0.5.1*

    Examples
    --------
    Consider a pipe-flow experiment, with tracers moving from side to side in
    multiple passes / streamlines. First locate the tracers, then split their
    trajectories into each individual pass:

    >>> import pept
    >>> from pept.tracking import *
    >>>
    >>> split_pipe = pept.Pipeline([
    >>>     Segregate(window = 10, max_distance = 20),  # Appends label column
    >>>     GroupBy("label"),                           # Splits into samples
    >>>     Reorient(),                                 # Align with X axis
    >>>     Center(),                                   # Center points at 0
    >>>     Stack(),
    >>> ])
    >>> streamlines = split_pipe.fit(trajectories)

    Now each sample in `streamlines` corresponds to a single tracer pass, e.g.
    `streamlines[0]` is the first pass, `streamlines[1]` is the second. The
    passes were reoriented and centred such that the pipe is aligned with the
    X axis.

    Now the `LaceyColors` reducer can be used to create an image of the mixing
    between the pipe entrance and exit:

    >>> from pept.processing import LaceyColors
    >>> entrance = [-100, 0, 0]     # Pipe data was aligned with X and centred
    >>> exit = [100, 0, 0]
    >>> lacey_image = LaceyColors(entrance, exit).fit(streamlines)
    >>> print(lacey_image.voxels)    # RGB channels of image
    (8, 8, 3)

    Now the image can be visualised e.g. with Plotly:

    >>> from pept.plots import PlotlyGrapher2D
    >>> PlotlyGrapher2D().add_image(lacey_image).show()
    '''

    def __init__(
        self,
        p1,
        p2,
        ax1 = None,
        ax2 = None,
        basis1 = None,
        basis2 = None,
        xlim = None,
        ylim = None,
        max_distance = 10,
        resolution = (8, 8),
    ):
        self.p1 = np.asarray(p1, dtype = float)
        _check_point3d(p1 = self.p1)

        if ax1 is not None:
            ax1 = np.asarray(ax1, dtype = float)
            ax1 = ax1 / np.linalg.norm(ax1)
            _check_point3d(ax1 = ax1)
        self.ax1 = ax1

        self.p2 = np.asarray(p2, dtype = float)
        _check_point3d(p2 = self.p2)

        if ax2 is not None:
            ax2 = np.asarray(ax2, dtype = float)
            ax2 = ax2 / np.linalg.norm(ax2)
            _check_point3d(ax2 = ax2)
        self.ax2 = ax2

        if basis1 is not None:
            basis1 = np.asarray(basis1, dtype = float)
            if basis1.ndim != 2 or len(basis1) != 3 or basis1.shape[1] != 2:
                raise ValueError(textwrap.fill((
                    "The input `basis1`, if defined, should be two column "
                    "vectors - i.e. a (3, 2) matrix. Received "
                    f"`basis1.shape={basis1.shape}`."
                )))
        self.basis1 = basis1

        if basis2 is not None:
            basis2 = np.asarray(basis2, dtype = float)
            if basis2.ndim != 2 or len(basis2) != 3 or basis2.shape[1] != 2:
                raise ValueError(textwrap.fill((
                    "The input `basis2`, if defined, should be two column "
                    "vectors - i.e. a (3, 2) matrix. Received "
                    f"`basis2.shape={basis2.shape}`."
                )))
        self.basis2 = basis2

        if xlim is not None:
            xlim = np.asarray(xlim, dtype = float)
            if xlim.ndim != 1 or len(xlim) != 2:
                raise ValueError(textwrap.fill((
                    "The input `xlim`, if given, should be a vector-like with "
                    f"two values [xmin, xmax]. Received `{xlim}`."
                )))
        self.xlim = xlim

        if ylim is not None:
            ylim = np.asarray(ylim, dtype = float)
            if ylim.ndim != 1 or len(ylim) != 2:
                raise ValueError(textwrap.fill((
                    "The input `ylim`, if given, should be a vector-like with "
                    f"two values [ymin, ymax]. Received `{ylim}`."
                )))
        self.ylim = ylim

        self.max_distance = float(max_distance)
        self.resolution = tuple(resolution)


    def fit(
        self,
        trajectories,
        executor = "joblib",
        max_workers = None,
        verbose = True,
    ):

        if not isinstance(trajectories, PointData):
            try:
                trajectories = PointData(trajectories)
            except (ValueError, TypeError):
                raise TypeError(textwrap.fill((
                    "The input `trajectories` must be pept.PointData-like. "
                    f"Received `{type(trajectories)}`."
                )))

        if self.ax1 is None:
            self.ax1 = self.p2 - self.p1

        if self.ax2 is None:
            self.ax2 = self.p2 - self.p1

        self.ax1 = self.ax1 / np.linalg.norm(self.ax1)
        self.ax2 = self.ax2 / np.linalg.norm(self.ax2)

        # Aliases
        p1 = self.p1
        ax1 = self.ax1
        p2 = self.p2
        ax2 = self.ax2

        max_distance = self.max_distance
        resolution = self.resolution

        # For each pass, find the closest point to the start and end planes
        points_start = []
        points_end = []

        for traj in trajectories:
            xyz = traj.points[:, 1:4]
            xyz = xyz[np.isfinite(xyz).all(axis = 1)]

            dists_start = np.abs(np.dot(xyz - p1, ax1))
            dists_end = np.abs(np.dot(xyz - p2, ax2))

            min_dist_start = np.min(dists_start)
            min_dist_end = np.min(dists_end)

            if min_dist_start < max_distance and min_dist_end < max_distance:
                points_start.append(xyz[np.argmin(dists_start)])
                points_end.append(xyz[np.argmin(dists_end)])

        points_start = np.array(points_start)
        points_end = np.array(points_end)

        # Split starting points into two species - Red and Blue - along the PCA
        points_start_centred = points_start - p1

        if self.basis1 is None:
            cov = np.cov(points_start_centred.T)
            evals, evecs = np.linalg.eig(cov)

            evals_argsorted = evals.argsort()[::-1]
            evals = evals[evals_argsorted]
            self.basis1 = evecs[:, evals_argsorted][:, :2]

        points_start2d = points_start_centred @ self.basis1

        # Project ending points onto 2D plane by the Principal Components
        points_end_centred = points_end - p2

        if self.basis2 is None:
            cov = np.cov(points_end_centred.T)
            evals, evecs = np.linalg.eig(cov)

            evals_argsorted = evals.argsort()[::-1]
            evals = evals[evals_argsorted]
            self.basis2 = evecs[:, evals_argsorted][:, :2]

        points_end2d = points_end_centred @ self.basis2

        # Create RGB image with the R and B channels == the number of passes
        ired = points_start2d[:, 0] < 0
        iblu = points_start2d[:, 0] >= 0

        # Compute image physical limits
        limits = np.c_[points_end2d.min(axis = 0), points_end2d.max(axis = 0)]

        if self.xlim is None:
            self.xlim = enlarge(limits[0, :])

        if self.ylim is None:
            self.ylim = enlarge(limits[1, :])

        xlim = self.xlim
        ylim = self.ylim

        xsize = (xlim[1] - xlim[0]) / resolution[0]
        ysize = (ylim[1] - ylim[0]) / resolution[1]

        red_mix = np.zeros(resolution)
        blu_mix = np.zeros(resolution)

        # Find indices within pixels
        red_ix = ((points_end2d[ired, 0] - xlim[0]) / xsize).astype(int)
        red_iy = ((points_end2d[ired, 1] - ylim[0]) / ysize).astype(int)

        # Ensure all red pixel indices are within bounds
        mask = (
            (red_ix >= 0) & (red_ix < resolution[0]) &
            (red_iy >= 0) & (red_iy < resolution[1])
        )

        red_ix = red_ix[mask]
        red_iy = red_iy[mask]

        for i in range(len(red_ix)):
            red_mix[red_ix[i], red_iy[i]] += 1

        blu_ix = ((points_end2d[iblu, 0] - xlim[0]) / xsize).astype(int)
        blu_iy = ((points_end2d[iblu, 1] - ylim[0]) / ysize).astype(int)

        # Ensure all blue pixel indices are within bounds
        mask = (
            (blu_ix >= 0) & (blu_ix < resolution[0]) &
            (blu_iy >= 0) & (blu_iy < resolution[1])
        )

        blu_ix = blu_ix[mask]
        blu_iy = blu_iy[mask]

        for i in range(len(blu_ix)):
            blu_mix[blu_ix[i], blu_iy[i]] += 1

        # Stack RGB channels
        red_mix *= 255 / red_mix.max()
        blu_mix *= 255 / blu_mix.max()

        mix_img = np.zeros((resolution[0], resolution[1], 3), dtype = int)
        mix_img[:, :, 0] = red_mix.astype(int)
        mix_img[:, :, 2] = blu_mix.astype(int)

        return Voxels(
            mix_img,
            xlim = xlim,
            ylim = ylim,
            zlim = [0, 2],
        )




class LaceyColorsLinear(Reducer):
    '''Apply the `LaceyColors` mixing algorithm at `num_divisions` equidistant
    points between `p1` and `p2`, saving images at each step in `directory`.

    Reducer signature:

    ::

               PointData -> LaceyColors.fit -> (height, width, 3) np.ndarray
         list[PointData] -> LaceyColors.fit -> (height, width, 3) np.ndarray
        list[np.ndarray] -> LaceyColors.fit -> (height, width, 3) np.ndarray

    For details about the mixing algorithm itself, check the `LaceyColors`
    documentation.

    The generated images (saved in `directory` with `height` x `width` pixels)
    can be stitched into a video using `pept.plots.make_video`.

    *New in pept-0.5.1*

    Examples
    --------
    Consider a pipe-flow experiment, with tracers moving from side to side in
    multiple passes / streamlines. First locate the tracers, then split their
    trajectories into each individual pass:

    >>> import pept
    >>> from pept.tracking import *
    >>>
    >>> split_pipe = pept.Pipeline([
    >>>     Segregate(window = 10, max_distance = 20),  # Appends label column
    >>>     GroupBy("label"),                           # Splits into samples
    >>>     Reorient(),                                 # Align with X axis
    >>>     Center(),                                   # Center points at 0
    >>>     Stack(),
    >>> ])
    >>> streamlines = split_pipe.fit(trajectories)

    Now each sample in `streamlines` corresponds to a single tracer pass, e.g.
    `streamlines[0]` is the first pass, `streamlines[1]` is the second. The
    passes were reoriented and centred such that the pipe is aligned with the
    X axis.

    Now the `LaceyColorsLinear` reducer can be used to create images of the
    mixing between the pipe entrance and exit:

    >>> from pept.processing import LaceyColorsLinear
    >>> entrance = [-100, 0, 0]     # Pipe data was aligned with X and centred
    >>> exit = [100, 0, 0]
    >>> LaceyColorsLinear(
    >>>     directory = "lacey",    # Creates directory and saves images there
    >>>     p1 = entrance,
    >>>     p2 = exit,
    >>> ).fit(streamlines)

    Now the directory "lacey" was created inside your current working folder,
    and all Lacey images saved there as "frame0000.png", "frame0001.png", etc.
    You can stitch all images together into a video using
    `pept.plots.make_video`:

    >>> import pept
    >>> pept.plots.make_video("lacey/frame*.png", output = "lacey/video.avi")
    '''

    def __init__(
        self,
        directory,
        p1, p2,
        xlim = None,
        ylim = None,
        num_divisions = 50,
        max_distance = 10,
        resolution = (8, 8),
        height = 1000,
        width = 1000,
        prefix = "frame",
    ):
        self.directory = directory
        self.num_divisions = int(num_divisions)
        if self.num_divisions <= 1:
            raise ValueError(textwrap.fill((
                "The input `num_divisions` should be an integer >= 2. "
                f"Received `{self.num_divisions}`."
            )))

        self.p1 = np.asarray(p1, dtype = float)
        self.p2 = np.asarray(p2, dtype = float)

        self.xlim = xlim
        self.ylim = ylim

        self.max_distance = float(max_distance)
        self.resolution = tuple(resolution)

        self.height = int(height)
        self.width = int(width)
        self.prefix = prefix


    def fit(
        self,
        trajectories,
        executor = "joblib",
        max_workers = None,
        verbose = True,
    ):

        if not os.path.isdir(self.directory):
            os.mkdir(self.directory)

        divisions = np.c_[
            np.linspace(self.p1[0], self.p2[0], self.num_divisions),
            np.linspace(self.p1[1], self.p2[1], self.num_divisions),
            np.linspace(self.p1[2], self.p2[2], self.num_divisions),
        ]

        axis = self.p2 - self.p1

        # Compute basis vectors for splitting flow
        lacey = LaceyColors(
            p1 = divisions[0],
            p2 = divisions[1],
            ax1 = axis,
            ax2 = axis,
            max_distance = self.max_distance,
            resolution = self.resolution,
        )
        lacey.fit(trajectories)

        basis1 = lacey.basis1
        basis2 = lacey.basis2

        if self.xlim is None:
            self.xlim = lacey.xlim

        if self.ylim is None:
            self.ylim = lacey.ylim

        def compute_save_lacey(idiv):
            image = LaceyColors(
                p1 = self.p1,
                p2 = divisions[idiv],
                ax1 = axis,
                ax2 = axis,
                basis1 = basis1,
                basis2 = basis2,
                xlim = self.xlim,
                ylim = self.ylim,
                max_distance = self.max_distance,
                resolution = self.resolution,
            ).fit(trajectories)

            travelled = np.linalg.norm(divisions[idiv] - self.p1)
            grapher = plots.PlotlyGrapher2D(subplot_titles = [
                f"Travelled {travelled} mm"
            ])
            grapher.add_image(image)
            grapher.fig.write_image(
                f"{self.directory}/{self.prefix}{idiv:0>4}.png",
                height = self.height,
                width = self.width,
            )

        _parallel(
            compute_save_lacey,
            range(1, self.num_divisions),
            executor = executor,
            max_workers = max_workers,
            desc = "LaceyColorsLinear :",
            verbose = verbose,
        )




class RelativeDeviations(Reducer):
    '''Compute a Lagrangian mixing measure - the changes in tracer distances
    to a point P1 as they pass through an "inlet" plane and another point P2
    when reaching an "outlet" plane.

    A deviation is computed for each tracer trajectory, yielding a range of
    deviations that can e.g be histogrammed (default). Intuitively, mixing is
    stronger if this distribution of deviations is wider.

    Reducer signature:

    ::

        If ``histogram = True`` (default)
               PointData -> LaceyColors.fit -> plotly.graph_objs.Figure
         list[PointData] -> LaceyColors.fit -> plotly.graph_objs.Figure
        list[np.ndarray] -> LaceyColors.fit -> plotly.graph_objs.Figure

        If ``histogram = False`` (return deviations)
               PointData -> LaceyColors.fit -> (N,) np.ndarray
         list[PointData] -> LaceyColors.fit -> (N,) np.ndarray
        list[np.ndarray] -> LaceyColors.fit -> (N,) np.ndarray


    **Each sample in the input `PointData` is treated as a separate streamline
    / tracer pass. You can group passes using `Segregate + GroupBy("label")`**.

    The first plane where the distances from tracers to a point `p1` is defined
    by the point `p1` and direction axis `ax1`. **The point `p1` should be the
    middle of the pipe**.

    The second plane where relative distances are evaluated is similarly
    defined by `p2` and `ax2`. **The point `p2` should be the middle of the
    volume / pipe**.

    If the direction vectors `ax1` and `ax2` are undefined (`None`), the
    tracers are assumed to follow a straight line between `p1` and `p2`.

    The `max_distance` parameter defines the maximum distance allowed between
    a point and a plane to be considered part of it. The `resolution` defines
    the number of pixels in the height and width of the resulting image.

    The following attributes are always set. A Plotly figure is only generated
    and returned if `histogram = True` (default).

    The extra keyword arguments ``**kwargs`` are passed to the histogram
    creation routine `pept.plots.histogram`. You can e.g. set the YAxis limits
    by adding `ylim = [0, 20]`.

    *New in pept-0.5.1*

    Attributes
    ----------
    points1 : pept.PointData
        The tracer points selected at the inlet around `p1`.

    points2 : pept.PointData
        The tracer points selected at the outlet around `p2`.

    deviations : (N,) np.ndarray
        The vector of tracer deviations for each tracer pass in `points1` and
        `points2`.

    Examples
    --------
    Consider a pipe-flow experiment, with tracers moving from side to side in
    multiple passes / streamlines. First locate the tracers, then split their
    trajectories into each individual pass:

    >>> import pept
    >>> from pept.tracking import *
    >>>
    >>> split_pipe = pept.Pipeline([
    >>>     Segregate(window = 10, max_distance = 20),  # Appends label column
    >>>     GroupBy("label"),                           # Splits into samples
    >>>     Reorient(),                                 # Align with X axis
    >>>     Center(),                                   # Center points at 0
    >>>     Stack(),
    >>> ])
    >>> streamlines = split_pipe.fit(trajectories)

    Now each sample in `streamlines` corresponds to a single tracer pass, e.g.
    `streamlines[0]` is the first pass, `streamlines[1]` is the second. The
    passes were reoriented and centred such that the pipe is aligned with the
    X axis.

    Now the `RelativeDeviations` reducer can be used to create a histogram of
    tracer deviations due to mixing:

    >>> from pept.processing import RelativeDeviations
    >>> entrance = [-100, 0, 0]     # Pipe data was aligned with X and centred
    >>> exit = [100, 0, 0]
    >>> fig = RelativeDeviations(entrance, exit).fit(streamlines)
    >>> fig.show()

    The deviations themselves can be extracted directly for further processing:

    >>> mixing_algorithm = RelativeDeviations(entrance, exit, histogram=False)
    >>> mixing_algorithm.fit(streamlines)

    >>> deviations = mixing_algorithm.deviations
    >>> inlet_points = mixing_algorithm.points1
    >>> outlet_points = mixing_algorithm.points2
    '''

    def __init__(
        self,
        p1,
        p2,
        ax1 = None,
        ax2 = None,
        max_distance = 10,
        histogram = True,
        **kwargs,
    ):
        self.p1 = np.asarray(p1, dtype = float)
        _check_point3d(p1 = self.p1)

        if ax1 is not None:
            ax1 = np.asarray(ax1, dtype = float)
            self.ax1 = ax1 / np.linalg.norm(ax1)
            _check_point3d(ax1 = self.ax1)

        self.p2 = np.asarray(p2, dtype = float)
        _check_point3d(p2 = self.p2)

        if ax2 is not None:
            ax2 = np.asarray(ax2, dtype = float)
            self.ax2 = ax2 / np.linalg.norm(ax2)
            _check_point3d(ax2 = self.ax2)

        self.max_distance = float(max_distance)
        self.histogram = bool(histogram)

        self.kwargs = kwargs

        # Will be set in `fit`
        self.points1 = None
        self.points2 = None
        self.deviations = None


    def fit(
        self,
        trajectories,
        executor = "joblib",
        max_workers = None,
        verbose = True,
    ):

        if not isinstance(trajectories, PointData):
            try:
                trajectories = PointData(trajectories)
            except (ValueError, TypeError):
                raise TypeError(textwrap.fill((
                    "The input `trajectories` must be pept.PointData-like. "
                    f"Received `{type(trajectories)}`."
                )))

        if self.ax1 is None:
            self.ax1 = self.p2 - self.p1

        if self.ax2 is None:
            self.ax2 = self.p2 - self.p1

        self.ax1 = self.ax1 / np.linalg.norm(self.ax1)
        self.ax2 = self.ax2 / np.linalg.norm(self.ax2)

        # Aliases
        p1 = self.p1
        ax1 = self.ax1
        p2 = self.p2
        ax2 = self.ax2

        max_distance = self.max_distance

        # For each pass, find the closest point to the start and end planes
        points_start = []
        points_end = []

        for traj in trajectories:
            xyz = traj.points[:, 1:4]
            xyz = xyz[np.isfinite(xyz).all(axis = 1)]

            dists_start = np.abs(np.dot(xyz - p1, ax1))
            dists_end = np.abs(np.dot(xyz - p2, ax2))

            min_dist_start = np.min(dists_start)
            min_dist_end = np.min(dists_end)

            if min_dist_start < max_distance and min_dist_end < max_distance:
                points_start.append(traj.points[np.argmin(dists_start)])
                points_end.append(traj.points[np.argmin(dists_end)])
            else:
                points_start.append(np.full(traj.points.shape[1], np.nan))
                points_end.append(np.full(traj.points.shape[1], np.nan))

        # TODO: check empty trajectories
        self.points1 = trajectories.copy(data = points_start)
        self.points2 = trajectories.copy(data = points_end)

        # Distances to P1 and P2 and relative deviations
        d1 = np.linalg.norm(self.points1.points[:, 1:4] - self.p1, axis = 1)
        d2 = np.linalg.norm(self.points2.points[:, 1:4] - self.p2, axis = 1)

        self.deviations = np.abs(d2 - d1)

        # Return histogram of deviations
        if self.histogram:
            return plots.histogram(
                self.deviations,
                xaxis_title = "Relative Deviation (mm)",
                yaxis_title = "Probability (%)",
                **self.kwargs,
            )
        else:
            return self.deviations




class RelativeDeviationsLinear(Reducer):
    '''Apply the `RelativeDeviations` mixing algorithm at `num_divisions`
    equidistant points between `p1` and `p2`, saving histogram images at each
    step in `directory`.

    Reducer signature:

    ::

               PointData -> LaceyColors.fit -> plotly.graph_objs.Figure
         list[PointData] -> LaceyColors.fit -> plotly.graph_objs.Figure
        list[np.ndarray] -> LaceyColors.fit -> plotly.graph_objs.Figure

    For details about the mixing algorithm itself, check the
    `RelativeDeviations` documentation.

    This algorithm saves a rich set of data:

    - Individual histograms for each point along P1-P2 are saved in the given
      `directory`.
    - A Plotly figure of computed statistics is returned, including the
      deviations' mean, standard deviation, skewness and kurtosis.
    - The raw data is saved as object attributes (see below).

    The generated images (saved in `directory` with `height` x `width` pixels)
    can be stitched into a video using `pept.plots.make_video`.

    The extra keyword arguments ``**kwargs`` are passed to the histogram
    creation routine `pept.plots.histogram`. You can e.g. set the YAxis limits
    by adding `ylim = [0, 20]`.

    *New in pept-0.5.1*

    Attributes
    ----------
    deviations : list[(N,) np.ndarray]
        A list of deviations computed by `RelativeDeviations` at each point
        between P1 and P2.

    mean : (N,) np.ndarray
        A vector of mean tracer deviations at each point between P1 and P2.

    std : (N,) np.ndarray
        A vector of the tracer deviations' standard deviation at each point
        between P1 and P2.

    skew : (N,) np.ndarray
        A vector of the tracer deviations' adjusted skewness at each point
        between P1 and P2. A normal distribution has a value of 0; positive
        values indicate a longer right distribution tail; negative values
        indicate a heavier left tail.

    kurtosis : (N,) np.ndarray
        A vector of the tracer deviations' Fisher kurtosis at each point
        between P1 and P2. A normal distribution has a value of 0; positive
        values indicate a "thin" distribution; negative values indicate a
        heavy, wide distribution.

    Examples
    --------
    Consider a pipe-flow experiment, with tracers moving from side to side in
    multiple passes / streamlines. First locate the tracers, then split their
    trajectories into each individual pass:

    >>> import pept
    >>> from pept.tracking import *
    >>>
    >>> split_pipe = pept.Pipeline([
    >>>     Segregate(window = 10, max_distance = 20),  # Appends label column
    >>>     GroupBy("label"),                           # Splits into samples
    >>>     Reorient(),                                 # Align with X axis
    >>>     Center(),                                   # Center points at 0
    >>>     Stack(),
    >>> ])
    >>> streamlines = split_pipe.fit(trajectories)

    Now each sample in `streamlines` corresponds to a single tracer pass, e.g.
    `streamlines[0]` is the first pass, `streamlines[1]` is the second. The
    passes were reoriented and centred such that the pipe is aligned with the
    X axis.

    Now the `RelativeDeviationsLinear` reducer can be used to create images of
    the mixing between the pipe entrance and exit:

    >>> from pept.processing import RelativeDeviationsLinear
    >>> entrance = [-100, 0, 0]     # Pipe data was aligned with X and centred
    >>> exit = [100, 0, 0]
    >>> summary_fig = RelativeDeviationsLinear(
    >>>     directory = "deviations",   # Creates directory to save images
    >>>     p1 = entrance,
    >>>     p2 = exit,
    >>> ).fit(streamlines)
    >>> summary_fig.show()              # Summary statistics: mean, std, etc.

    Now the directory "deviations" was created inside your current working
    folder, and all relative deviation histograms were saved there as
    "frame0000.png", "frame0001.png", etc.
    You can stitch all images together into a video using
    `pept.plots.make_video`:

    >>> import pept
    >>> pept.plots.make_video(
    >>>     "deviations/frame*.png",
    >>>     output = "deviations/video.avi"
    >>> )

    The raw deviations and statistics can also be extracted directly:

    >>> mixing_algorithm = RelativeDeviationsLinear(
    >>>     directory = "deviations",   # Creates directory to save images
    >>>     p1 = entrance,
    >>>     p2 = exit,
    >>> )
    >>> fig = mixing_algorithm.fit(streamlines)
    >>> fig.show()

    >>> deviations = mixing_algorithm.deviations
    >>> mean = mixing_algorithm.mean
    >>> std = mixing_algorithm.std
    >>> skew = mixing_algorithm.skew
    >>> kurtosis = mixing_algorithm.kurtosis
    '''

    def __init__(
        self,
        directory,
        p1, p2,
        num_divisions = 50,
        max_distance = 10,
        height = 1000,
        width = 2000,
        prefix = "frame",
        **kwargs,
    ):
        self.directory = directory
        self.num_divisions = int(num_divisions)

        self.p1 = np.asarray(p1, dtype = float)
        self.p2 = np.asarray(p2, dtype = float)

        self.max_distance = float(max_distance)

        self.height = int(height)
        self.width = int(width)
        self.prefix = prefix
        self.kwargs = kwargs

        # Will be set in `fit`
        self.deviations = None
        self.mean = None
        self.std = None
        self.skew = None
        self.kurtosis = None


    def fit(
        self,
        trajectories,
        executor = "joblib",
        max_workers = None,
        verbose = True,
    ):

        if not os.path.isdir(self.directory):
            os.mkdir(self.directory)

        divisions = np.c_[
            np.linspace(self.p1[0], self.p2[0], self.num_divisions),
            np.linspace(self.p1[1], self.p2[1], self.num_divisions),
            np.linspace(self.p1[2], self.p2[2], self.num_divisions),
        ]

        axis = self.p2 - self.p1

        # Save vectors of deviations so that we can set the XAxis range
        def compute_deviations(division):
            return RelativeDeviations(
                p1 = self.p1,
                p2 = division,
                ax1 = axis,
                ax2 = axis,
                max_distance = self.max_distance,
                histogram = False,
            ).fit(trajectories)

        self.deviations = _parallel(
            compute_deviations,
            divisions,
            executor = executor,
            max_workers = max_workers,
            desc = "RelativeDeviationsLinear 1 / 3 :",
            verbose = verbose,
        )

        # Save histograms to disk
        xlim = [0, np.nanmax([np.nanmax(d) for d in self.deviations])]

        def save_histograms(idiv):
            fig = plots.histogram(
                self.deviations[idiv],
                xlim = xlim,
                **self.kwargs,
            )

            travelled = np.linalg.norm(divisions[idiv] - self.p1)
            fig.update_layout(
                title = f"Travelled {travelled:4.4f} mm",
                xaxis_title = "Deviation (mm)",
                yaxis_title = "Probability (%)",
            )

            fig.write_image(
                f"{self.directory}/{self.prefix}{idiv:0>4}.png",
                height = self.height,
                width = self.width,
            )

        _parallel(
            save_histograms,
            range(self.num_divisions),
            executor = executor,
            max_workers = max_workers,
            desc = "RelativeDeviationsLinear 2 / 3 :",
            verbose = verbose,
        )

        # Compute summarised statistics about distributions
        def compute_stats(idiv):
            dev = self.deviations[idiv]
            dev = dev[np.isfinite(dev)]
            return [
                np.mean(dev),
                np.std(dev),
                scipy.stats.skew(dev),
                scipy.stats.kurtosis(dev),
            ]

        stats = _parallel(
            compute_stats,
            range(self.num_divisions),
            executor = executor,
            max_workers = max_workers,
            desc = "RelativeDeviationsLinear 3 / 3 :",
            verbose = verbose,
        )
        stats = np.array(stats, order = "F")

        self.mean = stats[:, 0]
        self.std = stats[:, 1]
        self.skew = stats[:, 2]
        self.kurtosis = stats[:, 3]

        # Plot summarised statistics
        distance = np.linspace(
            0,
            np.linalg.norm(self.p2 - self.p1),
            self.num_divisions,
        )

        fig = make_subplots(rows = 2, cols = 1, subplot_titles = [
            "Mean Tracer Deviation Along P1-P2 Axis",
            "Distribution Skewness & Kurtosis (How Many Outliers?)",
        ])

        fig.add_trace(
            go.Scatter(
                x = distance,
                y = self.mean,
                mode = "lines",
                name = "Mean Deviation",
            ),
            row = 1, col = 1,
        )

        fig.add_trace(
            go.Scatter(
                x = distance,
                y = self.mean - self.std,
                mode = 'lines',
                marker_color = "#444",
                line_width = 0,
                showlegend = False,
            ),
            row = 1, col = 1,
        )

        fig.add_trace(
            go.Scatter(
                x = distance,
                y = self.mean + self.std,
                mode = 'lines',
                marker_color = "#444",
                line_width = 0,
                fillcolor = 'rgba(68, 68, 68, 0.3)',
                fill = 'tonexty',
                name = "Standard Deviation",
            ),
            row = 1, col = 1,
        )

        # Plot skewness and kurtosis
        fig.add_trace(
            go.Scatter(
                x = distance,
                y = self.skew,
                name = "Skewness",
            ),
            row = 2, col = 1,
        )

        fig.add_trace(
            go.Scatter(
                x = distance,
                y = self.kurtosis,
                name = "Kurtosis",
            ),
            row = 2, col = 1,
        )

        plots.format_fig(fig)
        fig.update_xaxes(title = "Length Along P1-P2 Axis (mm)")
        fig.update_layout(yaxis_title = "Deviation (mm)")

        return fig




class AutoCorrelation(Reducer):
    r'''Compute autocorrelation of multiple measures (eg YZ velocities) as a
    function of a lagging variable (eg time).

    Reducer signature:

    ::

               PointData -> AutoCorrelation.fit -> PlotlyGrapher2D
         list[PointData] -> AutoCorrelation.fit -> PlotlyGrapher2D
        list[np.ndarray] -> AutoCorrelation.fit -> PlotlyGrapher2D

    **Each sample in the input `PointData` is treated as a separate streamline
    / tracer pass. You can group passes using `Segregate + GroupBy("label")`**.

    Autocorrelation and autocovariance each refer to about 3 different things
    in each field. The formula used here, inspired by the VACF in molecular
    dynamics and generalised for arbitrary measures, is:

    .. math::

       C(L_i) = \frac{ \sum_{N} V(L_0) \cdot V(L_i) }{N}

    i.e. the autocorrelation C at a lag of Li is the average of the dot
    products of quantities V for all N tracers. For example, the velocity
    autocorrelation function with respect to time would be the average of
    `vx(0) vx(t) + vy(0) vy(t) + vz(0) vz(t)` at a given time `t`.

    The input `lag` defines the column used as a lagging variable; it can be
    given as a named column string (e.g. `"t"`) or index (e.g. `0`).

    The input `signals` define the quantities for which the autocorrelation is
    computed, given as a list of column names (e.g. `["vy", "vz"]`) or indices
    (e.g. `[5, 6]`).

    The input `span`, if defined, is the minimum and maximum values for the
    `lag` (e.g. start and end times) for which the autocorrelation will be
    computed. By default it is automatically computed as the range of values.

    The input `num_divisions` is the number of lag points between `span[0]` and
    `span[1]` for which the autocorrelation will be computed.

    The `max_distance` parameter defines the maximum distance allowed between
    a lag value and the closest trajectory value for it to be considered.

    If `normalize` is True, then the formula used becomes:

    .. math::

       C(L_i) = \frac{ \sum_{N} V(L_0) \cdot V(L_i) / V(L_0) \cdot V(L_0)}{N}


    If `preprocess` is True, then the times of each tracer pass is taken
    relative to its start; only relevant if using time as the lagging variable.

    The extra keyword arguments ``**kwargs`` are passed to
    `PlotlyGrapher2D.add_points`. You can e.g. set the YAxis limits by adding
    `ylim = [0, 20]`.

    The extra keyword arguments ``**kwargs`` are passed to
    `plotly.graph_objs.Scatter`. You can e.g. set a different colorscheme with
    "marker_colorscheme = 'Viridis'".

    *New in pept-0.5.1*

    Examples
    --------
    Consider a pipe-flow experiment, with tracers moving from side to side in
    multiple passes / streamlines. First locate the tracers, then split their
    trajectories into each individual pass:

    >>> import pept
    >>> from pept.tracking import *
    >>>
    >>> split_pipe = pept.Pipeline([
    >>>     Segregate(window = 10, max_distance = 20),  # Appends label column
    >>>     GroupBy("label"),                           # Splits into samples
    >>>     Reorient(),                                 # Align with X axis
    >>>     Center(),                                   # Center points at 0
    >>>     Velocity(7),                                # Compute vx, vy, vz
    >>>     Stack(),
    >>> ])
    >>> streamlines = split_pipe.fit(trajectories)

    Now each sample in `streamlines` corresponds to a single tracer pass, e.g.
    `streamlines[0]` is the first pass, `streamlines[1]` is the second. The
    passes were reoriented and centred such that the pipe is aligned with the
    X axis.

    Now the `AutoCorrelation` algorithm can be used to compute the VACF:

    >>> from pept.processing import AutoCorrelation
    >>> fig = AutoCorrelation("t", ["vx", "vy", "vz"]).fit(streamlines)
    >>> fig.show()

    The radial velocity autocorrelation can be computed as a function of the
    pipe length (X axis as it was reoriented):

    >>> entrance = -100
    >>> exit = 100
    >>> ac = AutoCorrelation("x", ["vy", "vz"], span = [entrance, exit])
    >>> ac.fit(streamlines).show()

    The raw lags and autocorrelations plotted can be accessed directly:

    >>> ac.lags
    >>> ac.correlation

    The radial location can be autocorrelated with time, then normalised to
    show periodic movements (e.g. due to a mixer):

    >>> ac = AutoCorrelation("t", ["y", "z"], normalize = True)
    >>> ac.fit(streamlines).show()
    '''

    def __init__(
        self,
        lag = "t",
        signals = ["vx", "vy", "vz"],
        span = None,
        num_divisions = 500,
        max_distance = 10,
        normalize = False,
        preprocess = True,
        **kwargs,
    ):

        self.lag = lag
        if isinstance(signals, str) or isinstance(signals, Number):
            signals = [signals]
        self.signals = signals

        if span is not None:
            span = tuple(span)
            if len(span) != 2:
                raise ValueError(textwrap.fill((
                    "The input `span`, if defined, must contain two "
                    f"values [lag_start, lag_end]. Received `{span}`."
                )))

        self.span = span
        self.num_divisions = int(num_divisions)
        self.max_distance = float(max_distance)

        self.normalize = bool(normalize)
        self.preprocess = bool(preprocess)

        self.kwargs = kwargs

        # Will be set in `fit`
        self.lags = None
        self.correlation = None


    def fit(
        self,
        trajectories,
        executor = "joblib",
        max_workers = None,
        verbose = True,
    ):

        if not isinstance(trajectories, PointData):
            try:
                trajectories = PointData(trajectories)
            except (ValueError, TypeError):
                raise TypeError(textwrap.fill((
                    "The input `trajectories` must be pept.PointData-like. "
                    f"Received `{type(trajectories)}`."
                )))

        # Extract relevant columns' indices
        if isinstance(self.lag, str):
            ilag = trajectories.columns.index(self.lag)
        else:
            ilag = int(self.lag)

        isignals = []
        for sig in self.signals:
            if isinstance(sig, str):
                isignals.append(trajectories.columns.index(sig))
            else:
                isignals.append(int(sig))

        # If preprocessing is enabled, make times relative to the start of each
        # trajectory pass
        if self.preprocess and ilag == 0:
            trajectories = trajectories.copy()
            for traj in trajectories:
                traj.points[:, 0] -= traj.points[0, 0]

        # Set span if undefined
        if self.span is None:
            self.span = (
                np.min(trajectories.points[:, ilag]),
                np.max(trajectories.points[:, ilag]),
            )

        # For each tracer pass, find closest value to the lag in `divisions`
        divisions = np.linspace(self.span[0], self.span[1], self.num_divisions)
        correlation = np.zeros(self.num_divisions)

        # Save correlations as class attributes
        self.lags = divisions
        self.correlation = correlation

        def closest_signal(traj, lag):
            lag_values = traj.points[:, ilag]

            # Find closest lag value
            lag_distances = np.abs(lag_values - lag)
            ilag_closest = np.argmin(lag_distances)

            # If not lag was close enough, return NaNs
            if lag_distances[ilag_closest] > self.max_distance:
                return np.full(len(isignals), np.nan)
            else:
                return traj.points[ilag_closest, isignals]

        # Compute signals at lag=0
        signals0 = [
            closest_signal(traj, divisions[0])
            for traj in trajectories
        ]

        def compute_correlation(lag):
            # Find signal values close to this lag value
            signal_values = [
                closest_signal(traj, lag)
                for traj in trajectories
            ]

            # Compute correlation as dot product with signals at lag=0
            if self.normalize:
                corr = [
                    np.dot(signals0[i], signal_values[i]) /
                    np.dot(signals0[i], signals0[i])
                    for i in range(len(signal_values))
                ]
            else:
                corr = [
                    np.dot(signals0[i], signal_values[i])
                    for i in range(len(signal_values))
                ]

            # Return ensemble average, ignoring NaNs (i.e. when no signal was
            # found close enough to the given lag value)
            finites = np.isfinite(corr)

            if not len(corr) or np.all(~finites):
                return np.nan, 0

            return np.nanmean(corr), finites.sum()

        # Compute correlation at each lag in `divisions`
        stats = _parallel(
            compute_correlation,
            divisions,
            executor = executor,
            max_workers = max_workers,
            desc = "AutoCorrelation :",
            verbose = verbose,
        )

        correlation[:] = [s[0] for s in stats]
        finites = np.array([s[1] for s in stats])

        # Make points denser for colorbars
        dense_lags = np.linspace(
            self.span[0],
            self.span[1],
            50 * self.num_divisions,
        )
        dense_correlation = interp1d(self.lags, self.correlation)(dense_lags)
        dense_finites = interp1d(self.lags, finites)(dense_lags)

        # Plot correlation as a function of lag
        if len(self.correlation):
            ylim = [np.nanmin(self.correlation), np.nanmax(self.correlation)]
        else:
            ylim = None

        if "marker_size" not in self.kwargs.keys():
            self.kwargs["marker_size"] = 5

        if "marker_colorscale" not in self.kwargs.keys():
            self.kwargs["marker_colorscale"] = "Magma_r"

        if "line_color" not in self.kwargs.keys():
            self.kwargs["line_color"] = "black"

        grapher = plots.PlotlyGrapher2D(ylim = ylim)
        grapher.add_trace(go.Scatter(
            x = dense_lags,
            y = dense_correlation,
            mode = "markers+lines",
            marker = dict(
                colorbar_title = "N",
                color = dense_finites,
            ),
            **self.kwargs,
        ))

        norm = "Normalised " if self.normalize else ""
        grapher.fig.update_layout(
            xaxis_title = f"Lag (column `{self.lag}`)",
            yaxis_title = f"{norm}AutoCorrelation (columns {self.signals})",
        )

        # Set scaleratios to be automatically determined
        for i in range(grapher._rows):
            for j in range(grapher._cols):
                index = i * grapher._cols + j + 1
                yaxis = f"yaxis{index}" if index != 1 else "yaxis"
                grapher._fig["layout"][yaxis].update(
                    scaleanchor = None,
                    scaleratio = None,
                )

        # If the span is inversed, reverse the X-axis
        if self.span[1] < self.span[0]:
            grapher.fig.update_layout(xaxis_autorange = "reversed")

        return grapher




class SpatialProjections(Reducer):
    '''Project multiple tracer passes onto a moving 2D plane along a given
    `direction` between `start` and `end` coordinates, saving each frame in
    `directory`.

    Reducer signature:

    ::

               PointData -> SpatialProjections.fit -> None
         list[PointData] -> SpatialProjections.fit -> None
        list[np.ndarray] -> SpatialProjections.fit -> None

    **Each sample in the input `PointData` is treated as a separate streamline
    / tracer pass. You can group passes using `Segregate + GroupBy("label")`**.

    The generated images (saved in `directory` with `height` x `width` pixels)
    can be stitched into a video using `pept.plots.make_video`.

    The extra keyword arguments ``**kwargs`` are passed to the histogram
    creation routine `pept.plots.histogram`. You can e.g. set the YAxis limits
    by adding `ylim = [0, 20]`.

    *New in pept-0.5.1*

    Attributes
    ----------
    projections : list[(N, 5), np.ndarray]
        A list of frames for each division between `start` and `end`, with each
        frame saving 5 columns [t, x, y, z, colorbar_col].

    Examples
    --------
    Consider a pipe-flow experiment, with tracers moving from side to side in
    multiple passes / streamlines. First locate the tracers, then split their
    trajectories into each individual pass:

    >>> import pept
    >>> from pept.tracking import *
    >>>
    >>> split_pipe = pept.Pipeline([
    >>>     Segregate(window = 10, max_distance = 20),  # Appends label column
    >>>     GroupBy("label"),                           # Splits into samples
    >>>     Reorient(),                                 # Align with X axis
    >>>     Center(),                                   # Center points at 0
    >>>     Stack(),
    >>> ])
    >>> streamlines = split_pipe.fit(trajectories)

    Now each sample in `streamlines` corresponds to a single tracer pass, e.g.
    `streamlines[0]` is the first pass, `streamlines[1]` is the second. The
    passes were reoriented and centred such that the pipe is aligned with the
    X axis.

    Now the `RelativeDeviationsLinear` reducer can be used to create images of
    the mixing between the pipe entrance and exit:

    >>> from pept.processing import SpatialProjections
    >>> entrance_x = -100                   # Pipe data was aligned with X
    >>> exit_x = 100
    >>> SpatialProjections(
    >>>     directory = "projections",      # Creates directory to save images
    >>>     start = entrance_x,
    >>>     end = exit_x,
    >>> ).fit(streamlines)

    Now the directory "projections" was created inside your current working
    folder, and eachc projected frame was saved there as "frame0000.png",
    "frame0001.png", etc. You can stitch all images together into a video using
    `pept.plots.make_video`:

    >>> import pept
    >>> pept.plots.make_video(
    >>>     "projections/frame*.png",
    >>>     output = "projections/video.avi"
    >>> )

    The raw projections can also be extracted directly:

    >>> sp = SpatialProjections(
    >>>     directory = "projections",   # Creates directory to save images
    >>>     p1 = entrance_x,
    >>>     p2 = exit_x,
    >>> )
    >>> sp.fit(streamlines)
    >>> sp.projections
    '''

    def __init__(
        self,
        directory,
        start,
        end,
        dimension = "x",
        num_divisions = 500,
        max_distance = 10,
        colorbar_col = -1,
        height = 1000,
        width = 1000,
        prefix = "frame",
        **kwargs,
    ):

        self.directory = directory
        self.start = float(start)
        self.end = float(end)

        # Transform x -> 1, y -> 2, z -> 3 using ASCII integer `ord`er
        if isinstance(dimension, str):
            if dimension != "x" and dimension != "y" and dimension != "z":
                raise ValueError(textwrap.fill((
                    "The input `dimension` must be either 'x', 'y' or 'z'. "
                    f"Received `{dimension}`."
                )))

            self.dimension = ord(dimension) - ord('x') + 1
        else:
            self.dimension = int(dimension)
            if self.dimension < 1 or self.dimension > 3:
                raise ValueError(textwrap.fill((
                    "The input `dimension`, if given as an index, must be "
                    f"between 1 and 3 (inclusive). Received `{dimension}`."
                )))

        self.num_divisions = int(num_divisions)
        self.max_distance = float(max_distance)
        self.colorbar_col = colorbar_col

        self.height = int(height)
        self.width = int(width)
        self.prefix = prefix
        self.kwargs = kwargs

        # Will be set in `fit`
        self.projections = None


    def fit(
        self,
        trajectories,
        executor = "joblib",
        max_workers = None,
        verbose = True,
    ):

        if not isinstance(trajectories, PointData):
            try:
                trajectories = PointData(trajectories)
            except (ValueError, TypeError):
                raise TypeError(textwrap.fill((
                    "The input `trajectories` must be pept.PointData-like. "
                    f"Received `{type(trajectories)}`."
                )))

        # Find colorbar column's index
        if isinstance(self.colorbar_col, str):
            icolorbar_col = trajectories.columns.index(self.colorbar_col)
        else:
            icolorbar_col = int(self.colorbar_col)

        # Create directory for saving projection plots
        if not os.path.isdir(self.directory):
            os.mkdir(self.directory)

        # Points at which to project the trajectories
        self.divisions = np.linspace(self.start, self.end, self.num_divisions)

        # For each point along the start-end axis, save the closest point's
        # coordinates and colorbar column
        def project_division(idiv):
            division = self.divisions[idiv]
            points = []

            for traj in trajectories:
                coords = traj.points[:, self.dimension]
                distances = np.abs(coords - division)

                imin_dist = np.argmin(distances)
                min_dist = distances[imin_dist]

                if min_dist < self.max_distance:
                    points.append(traj.points[
                        imin_dist,
                        [0, 1, 2, 3, icolorbar_col],
                    ])

            if not len(points):
                return np.empty((0, 5))
            return np.array(points)

        self.projections = _parallel(
            project_division,
            range(self.num_divisions),
            executor = executor,
            max_workers = max_workers,
            desc = "SpatialProjections 1 / 2 :",
            verbose = verbose,
        )

        # Other remaining dimensions
        other = list({1, 2, 3} - {self.dimension})
        ix = other[0]
        iy = other[1]

        # Find plot limits
        if "xlim" not in self.kwargs.keys():
            points = np.concatenate([p[:, ix] for p in self.projections])
            xlim = [points.min(), points.max()]
            extra = 0.05 * (xlim[1] - xlim[0])
            self.kwargs["xlim"] = [xlim[0] - extra, xlim[1] + extra]

        if "ylim" not in self.kwargs.keys():
            points = np.concatenate([p[:, iy] for p in self.projections])
            ylim = [points.min(), points.max()]
            extra = 0.05 * (ylim[1] - ylim[0])
            self.kwargs["ylim"] = [ylim[0] - extra, ylim[1] + extra]

        if "size" not in self.kwargs.keys():
            self.kwargs["size"] = 15

        colors = np.concatenate([p[:, -1] for p in self.projections])
        if "marker_cmin" not in self.kwargs.keys():
            self.kwargs["marker_cmin"] = colors.min()

        if "marker_cmax" not in self.kwargs.keys():
            self.kwargs["marker_cmax"] = colors.max()

        if "colorscale" not in self.kwargs.keys():
            self.kwargs["marker_colorscale"] = "Magma"

        if "colorbar_title" not in self.kwargs.keys():
            self.kwargs["marker_colorbar_title"] = \
                trajectories.columns[icolorbar_col]

        def plot_save_projections(idiv):
            # dim = '_XYZ'[self.dimension]
            # dist = self.divisions[idiv] - self.divisions[0]

            grapher = plots.PlotlyGrapher2D(
                # subplot_titles = [f"Length along {dim} = {dist:4.4f} mm"],
                xlim = self.kwargs["xlim"],
                ylim = self.kwargs["ylim"],
            )

            kwargs = self.kwargs.copy()
            del kwargs["xlim"]
            del kwargs["ylim"]

            grapher.add_points(
                self.projections[idiv][:, [0, ix, iy]],
                color = self.projections[idiv][:, -1],
                **kwargs,
            )

            # Set axis labels
            grapher.fig.update_layout(
                xaxis_title = f"{'_XYZ'[other[0]]} (mm)",
                yaxis_title = f"{'_XYZ'[other[1]]} (mm)",
            )

            grapher.fig.write_image(
                f"{self.directory}/{self.prefix}{idiv:0>4}.png",
                height = self.height,
                width = self.width,
            )

        self.projections = _parallel(
            plot_save_projections,
            range(self.num_divisions),
            executor = executor,
            max_workers = max_workers,
            desc = "SpatialProjections 2 / 2 :",
            verbose = verbose,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : space_transformers.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 11.08.2021


import  textwrap

import  numpy               as      np
from    scipy.interpolate   import  interp1d
from    scipy.spatial       import  KDTree

import  pept
from    pept                import  LineData, PointData, Voxels
from    pept.base           import  LineDataFilter, PointDataFilter, Reducer




class Voxelize(LineDataFilter):
    '''Asynchronously voxelize samples of lines from a `pept.LineData`.

    Filter signature:

    ::

        LineData -> Voxelize.fit_sample -> PointData

    This filter is much more memory-efficient than voxelizing all samples of
    LoRs at once - which often overflows the available memory. Most often this
    is used alongside voxel-based tracking algorithms, e.g.
    ``pept.tracking.FPI``:

    >>> from pept.tracking import *
    >>> pipeline = pept.Pipeline([
    >>>     Voxelize((50, 50, 50)),
    >>>     FPI(3, 0.4),
    >>>     Stack(),
    >>> ])

    Parameters
    ----------
    number_of_voxels : 3-tuple
        A tuple-like containing exactly three integers specifying the number of
        voxels to be used in each dimension.

    xlim : (2,) list[float], optional
        The lower and upper boundaries of the voxellised volume in the
        x-dimension, formatted as [x_min, x_max]. If undefined, it is
        inferred from the bounding box of each sample of lines.

    ylim : (2,) list[float], optional
        The lower and upper boundaries of the voxellised volume in the
        y-dimension, formatted as [y_min, y_max]. If undefined, it is
        inferred from the bounding box of each sample of lines.

    zlim : (2,) list[float], optional
        The lower and upper boundaries of the voxellised volume in the
        z-dimension, formatted as [z_min, z_max]. If undefined, it is
        inferred from the bounding box of each sample of lines.

    set_lims : (N, 7) numpy.ndarray or pept.LineData, optional
        If defined, set the system limits upon creating the class to the
        bounding box of the lines in `set_lims`.
    '''

    def __init__(
        self,
        number_of_voxels,
        xlim = None,
        ylim = None,
        zlim = None,
        set_lims = None,
    ):
        # Type-checking inputs
        number_of_voxels = np.asarray(
            number_of_voxels,
            order = "C",
            dtype = int
        )

        if number_of_voxels.ndim != 1 or len(number_of_voxels) != 3:
            raise ValueError(textwrap.fill((
                "The input `number_of_voxels` must be a list-like "
                "with exactly three values, corresponding to the "
                "number of voxels in the x-, y-, and z-dimension. "
                f"Received parameter with shape {number_of_voxels.shape}."
            )))

        if (number_of_voxels < 2).any():
            raise ValueError(textwrap.fill((
                "The input `number_of_voxels` must set at least two "
                "voxels in each dimension (i.e. all elements in "
                "`number_of_elements` must be larger or equal to two). "
                f"Received `{number_of_voxels}`."
            )))

        # Keep track of limits we have to set
        set_xlim = True
        set_ylim = True
        set_zlim = True

        if xlim is not None:
            set_xlim = False
            xlim = np.asarray(xlim, dtype = float)

            if xlim.ndim != 1 or len(xlim) != 2:
                raise ValueError(textwrap.fill((
                    "The input `xlim` parameter must be a list with exactly "
                    "two values, corresponding to the minimum and maximum "
                    "coordinates of the voxel space in the x-dimension. "
                    f"Received parameter with shape {xlim.shape}."
                )))

        if ylim is not None:
            set_ylim = False
            ylim = np.asarray(ylim, dtype = float)

            if ylim.ndim != 1 or len(ylim) != 2:
                raise ValueError(textwrap.fill((
                    "The input `ylim` parameter must be a list with exactly "
                    "two values, corresponding to the minimum and maximum "
                    "coordinates of the voxel space in the y-dimension. "
                    f"Received parameter with shape {ylim.shape}."
                )))

        if zlim is not None:
            set_zlim = False
            zlim = np.asarray(zlim, dtype = float)

            if zlim.ndim != 1 or len(zlim) != 2:
                raise ValueError(textwrap.fill((
                    "The input `zlim` parameter must be a list with exactly "
                    "two values, corresponding to the minimum and maximum "
                    "coordinates of the voxel space in the z-dimension. "
                    f"Received parameter with shape {ylim.shape}."
                )))

        # Setting class attributes
        self._number_of_voxels = tuple(number_of_voxels)
        self._xlim = xlim
        self._ylim = ylim
        self._zlim = zlim

        if set_lims is not None:
            if isinstance(set_lims, LineData):
                self.set_lims(set_lims.lines, set_xlim, set_ylim, set_zlim)
            else:
                self.set_lims(set_lims, set_xlim, set_ylim, set_zlim)


    def set_lims(
        self,
        lines,
        set_xlim = True,
        set_ylim = True,
        set_zlim = True,
    ):
        lines = np.asarray(lines, dtype = float, order = "C")

        if set_xlim:
            xlim = Voxels.get_cutoff(lines[:, 1], lines[:, 4])

        if set_ylim:
            ylim = Voxels.get_cutoff(lines[:, 2], lines[:, 5])

        if set_zlim:
            zlim = Voxels.get_cutoff(lines[:, 3], lines[:, 6])

        self._xlim = xlim
        self._ylim = ylim
        self._zlim = zlim


    @property
    def number_of_voxels(self):
        return self._number_of_voxels


    @property
    def xlim(self):
        return self._xlim


    @property
    def ylim(self):
        return self._ylim


    @property
    def zlim(self):
        return self._zlim


    def fit_sample(self, sample_lines):
        if not isinstance(sample_lines, LineData):
            sample_lines = LineData(sample_lines)

        vox = Voxels.from_lines(
            sample_lines.lines,
            self.number_of_voxels,
            self.xlim,
            self.ylim,
            self.zlim,
            verbose = False,
        )

        # Save the constituent LoRs as a hidden attribute
        vox.attrs["_lines"] = sample_lines
        return vox




class Interpolate(PointDataFilter):
    '''Interpolate between data points at a fixed sampling rate; useful for
    Eulerian fields computation.

    Filter signature:

    ::

        PointData -> Interpolate.fit_sample -> PointData

    By default, the linear interpolator `scipy.interpolate.interp1d` is used.
    You can specify a different interpolator and keyword arguments to send it.
    E.g. nearest-neighbour interpolation: ``Interpolate(1., kind = "nearest")``
    or cubic interpolation: ``Interpolate(1., kind = "cubic")``.

    All data columns except timestamps are interpolated.
    '''

    def __init__(self, timestep, interpolator = interp1d, **kwargs):
        self.timestep = float(timestep)
        self.interpolator = interpolator
        self.kwargs = kwargs


    def fit_sample(self, sample):
        if not isinstance(sample, PointData):
            sample = PointData(sample)

        if not len(sample):
            return sample[0:0]

        # Create interpolators for each dimension (column) wrt time
        interps = [
            self.interpolator(
                sample.points[:, 0],
                sample.points[:, i],
                **self.kwargs,
            ) for i in range(1, sample.points.shape[1])
        ]

        # Sampling points, between the first and last timestamps
        sampling = np.arange(sample.points[0, 0], sample.points[-1, 0],
                             self.timestep)

        data = np.vstack([sampling] + [interp(sampling) for interp in interps])
        return sample.copy(data = data.T)




class Reorient(Reducer):
    '''Rotate a dataset such that it is oriented according to its principal
    axes.

    Reducer signature:

    ::

              PointData -> Reorient.fit -> PointData
        list[PointData] -> Reorient.fit -> PointData
             np.ndarray -> Reorient.fit -> PointData

    By default, this reducer reorients the points such that the axis along
    which it is most spread out (e.g. lengthwise in a pipe) becomes the X-axis.
    The input argument `dimensions` sets this - the default `"xyz"` can be
    changed to e.g. `"zyx"` so that the longest data axis becomes the Z-axis.

    The reducer also sets three attributes on the returned `PointData`:
    - `origin`: the origin relative to which the initial data was rotated.
    - `basis`: the principal components - or change of basis 3x3 matrix.
    - `eigenvalues`: how spread out the data is in each initial dimension.

    If you'd like to reorient a second dataset to the same basis as a first
    one, set the `basis` and `origin` arguments.

    *New in pept-0.5.0*

    Examples
    --------
    Reorient a dataset by aligning the longest principal component (e.g.
    lengthwise in a pipe) to the X-axis:

    >>> import pept.tracking as pt
    >>> data = PointData(...)
    >>> reoriented = pt.Reorient().fit(data)

    Reorient it such that the longest principal component (e.g. vertical in a
    mixer) becomes the Z-axis:

    >>> reoriented = pt.Reorient("zyx").fit(data)

    Reorient a second dataset to the same orientation basis as the first one:

    >>> reoriented2 = pt.Reorient(
    >>>     basis = reoriented.attrs["basis"],
    >>>     origin = reoriented.attrs["origin"],
    >>> ).fit(other_data)
    '''

    def __init__(self, dimensions = "xyz", basis = None, origin = None):
        # Type-checking inputs
        self.dimensions = np.array([0, 0, 0])
        if isinstance(dimensions, str):
            d = str(dimensions)
            if len(d) != 3 or d[0] not in "xyz" or d[1] not in "xyz" or \
                    d[2] not in "xyz":
                raise ValueError(textwrap.fill((
                    "The input `dimensions`, if given as a str, must have "
                    "exactly three characters containing 'x', 'y', or 'z'; "
                    f"e.g. 'xyz', 'zxy'. Received `{d}`."
                )))

            # Transform x -> 1, y -> 2, z -> 3 using ASCII integer `ord`er
            self.dimensions[0] = ord(d[0]) - ord('x') + 1
            self.dimensions[1] = ord(d[1]) - ord('x') + 1
            self.dimensions[2] = ord(d[2]) - ord('x') + 1
        else:
            d = np.asarray(dimensions, dtype = int)
            if d.ndim != 1 or len(d) != 3:
                raise ValueError(textwrap.fill((
                    "The input `dimensions`, if given as a list, must contain "
                    "exactly three integers representing the column indices "
                    "to use; e.g. `[1, 2, 3]` for xyz, `[3, 1, 2]` for zxy. "
                    f"Received `{d}`."
                )))
            self.dimensions[0] = d[0]
            self.dimensions[1] = d[1]
            self.dimensions[2] = d[2]

        if (basis is None and origin is not None) or \
                (basis is not None and origin is None):
            raise ValueError(textwrap.fill((
                "If a change of `basis` matrix is given, an `origin` is "
                "required too to do the rotation relative to."
            )))

        if basis is not None:
            basis = np.asarray(basis, dtype = float)
            if basis.ndim != 2 or basis.shape != (3, 3):
                raise ValueError(textwrap.fill((
                    "The input `basis`, if defined, must be a 3x3 matrix. "
                    f"Received `{basis.shape}` matrix."
                )))

        if origin is not None:
            origin = np.asarray(origin, dtype = float)
            if origin.ndim != 1 or basis.shape[0] != 3:
                raise ValueError(textwrap.fill((
                    "The input `origin`, if defined, must be a (3,) vector. "
                    f"Received `{origin.shape}` matrix."
                )))

        self.basis = basis
        self.origin = origin


    def fit(self, samples):
        # Reduce / stack list of samples onto a single PointData / array
        samples = pept.tracking.Stack().fit(samples)

        if not isinstance(samples, PointData):
            samples = PointData(samples)

        points = samples.points[:, 1:4]

        # Compute principal components - i.e. eigenvectors of covariance matrix
        if self.basis is None:
            points_mean = points.mean(axis = 0)
            points_centred = points - points_mean

            cov = np.cov(points_centred.T)
            evals, evecs = np.linalg.eig(cov)

            # Order eigenvectors such that the largest eigenvalue (i.e. the
            # dimensions that's most spread out) corresponds to the first of
            # `self.dimesions`
            reorder = evals.argsort()[2 - self.dimensions.argsort()]
            eigenvalues = evals[reorder]
            basis = evecs[:, reorder]

        else:
            points_mean = self.origin
            points_centred = points - points_mean
            basis = self.basis
            eigenvalues = None

        rotpoints = samples.points.copy()
        rotpoints[:, 1:4] = points_mean + points_centred @ basis

        return samples.copy(
            data = rotpoints,
            sample_size = samples.sample_size,
            overlap = samples.overlap,
            eigenvalues = eigenvalues,
            basis = basis,
            origin = points_mean,
        )




class Center(pept.base.Reducer):
    '''Center a dataset around `origin` (default [0, 0, 0]).

    Reducer signature:

    ::

              PointData -> Center.fit -> PointData
        list[PointData] -> Center.fit -> PointData
             np.ndarray -> Center.fit -> PointData

    *New in pept-0.5.1*

    '''

    def __init__(self, origin = None):
        if origin is not None:
            self.origin = np.asarray(origin)
            if self.origin.ndim != 1 or len(self.origin) != 3:
                raise ValueError(textwrap.fill((
                    "The input `origin` must be a 1D vector-like with exactly "
                    f"3 elements [x, y, z]. Received {self.origin}."
                )))
        else:
            self.origin = origin


    def fit(self, sample):
        sample = pept.tracking.Stack().fit(sample)
        if not isinstance(sample, pept.PointData):
            sample = pept.PointData(sample)

        if self.origin is None:
            self.origin = np.nanmedian(sample.points[:, 1:4], axis = 0)

        centered = sample.points.copy()
        centered[:, 1:4] -= self.origin

        return sample.copy(
            data = centered,
            sample_size = sample.sample_size,
            overlap = sample.overlap,
        )




class OutOfViewFilter(pept.base.Reducer):
    '''Remove tracer locations that are sparse *in time* - ie the ``k``-th
    nearest detection is later than ``max_time``.

    Reducer signature:

    ::

              PointData -> OutOfViewFilter.fit -> PointData
        list[PointData] -> OutOfViewFilter.fit -> PointData
          numpy.ndarray -> OutOfViewFilter.fit -> PointData

    This reducer (i.e. stacks all data samples, then processes it) is useful
    when the tracer goes out of the PEPT scanners and there are a few sparse
    noisy detections to remove.

    *New in pept-0.5.1*

    Examples
    --------
    Select only tracer locations whose next detection is within 200 ms.

    >>> import pept
    >>> import pept.tracking as pt
    >>> trajectories = pept.PointData(...)
    >>> # Only keep points whose next detection is within 200 ms
    >>> inview = pt.OutOfViewFilter(max_time = 200.).fit(trajectories)

    '''

    def __init__(self, max_time = 200., k = 5):
        self.max_time = float(max_time)
        self.k = int(k)


    def fit(self, samples):
        # Reduce / stack list of samples onto a single PointData / array
        samples = pept.tracking.Stack().fit(samples)

        if not isinstance(samples, pept.base.PEPTObject):
            samples = pept.PointData(samples)

        # Sort points by time
        points = samples.data[samples.data[:, 0].argsort()]

        # Select only points whose k-th nearest detection is sooner than
        # `max_time`
        times = points[:, 0][:, None]
        tree = KDTree(times)
        dt, _ = tree.query(times, [self.k + 1])

        nonsparse = points[dt[:, 0] < self.max_time]
        return samples.copy(data = nonsparse)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : transformers.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 08.08.2021


import  re
import  warnings
from    numbers         import  Number
import  textwrap

import  numpy           as      np
import  pandas          as      pd
from    scipy.optimize  import  minimize_scalar

from    pept.base       import  PointData, LineData
from    pept.base       import  Filter, Reducer, IterableSamples
from    pept.base       import  PEPTObject, AdaptiveWindow




class Stack(Reducer):
    '''Stack iterables - for example a ``list[pept.LineData]`` into a single
    ``pept.LineData``, a ``list[list]`` into a flattened ``list``.

    Reducer signature:

    ::

             list[LineData] -> Stack.fit -> LineData
            list[PointData] -> Stack.fit -> PointData

            list[list[Any]] -> Stack.fit -> list[Any]
        list[numpy.ndarray] -> Stack.fit -> numpy.ndarray

                      other -> Stack.fit -> other

    Can optionally set a given `sample_size` and `overlap`. This is useful
    when collecting a list of processed samples back into a single object.
    '''

    def __init__(self, sample_size = None, overlap = None):
        self.sample_size = sample_size
        self.overlap = overlap


    def fit(self, samples):
        # If it's a LineData / PointData, the `samples` are already stacked.
        # Simply set the sample_size and overlap if required and return them
        if not hasattr(samples, "__iter__"):
            raise ValueError(textwrap.fill((
                "The input `samples` must be an iterable (e.g. list, tuple, "
                f"PointData, LineData). Received type=`{type(samples)}`."
            )))

        if isinstance(samples, IterableSamples):
            if self.sample_size is not None:
                samples.sample_size = self.sample_size
            if self.sample_size is not None:
                samples.overlap = self.overlap
            return samples

        # If it's an empty iterator, we don't have anything to stack
        if len(samples) == 0:
            return samples

        # Stack Lines into LineData
        elif isinstance(samples[0], LineData):
            samples = LineData(samples)

        # Stack Points into PointData
        elif isinstance(samples[0], PointData):
            samples = PointData(samples)

        # Flatten list of lists
        elif isinstance(samples[0], list):
            samples = [item for sublist in samples for item in sublist]

        # Vertically stack list of NumPy arrays
        elif isinstance(samples[0], np.ndarray):
            samples = np.vstack(samples)

        # Set new sample_size and overlap if required
        if self.sample_size is not None:
            samples.sample_size = self.sample_size
        if self.sample_size is not None:
            samples.overlap = self.overlap

        return samples




class SplitLabels(Filter):
    '''Split a sample of data into unique ``label`` values, optionally removing
    noise and extracting `_lines` attributes.

    Filter signature:

    ::

        # `extract_lines` = False (default)
         LineData -> SplitLabels.fit_sample -> list[LineData]
        PointData -> SplitLabels.fit_sample -> list[PointData]

        # `extract_lines` = True and PointData.attrs["_lines"] exists
        PointData -> SplitLabels.fit_sample -> list[LineData]


    The sample of data must have a column named exactly "label". If
    ``remove_label = True`` (default), the "label" column is removed.
    '''

    def __init__(
        self,
        remove_labels = True,
        extract_lines = False,
        noise = False,
    ):
        self.remove_labels = bool(remove_labels)
        self.extract_lines = bool(extract_lines)
        self.noise = bool(noise)


    def _get_cluster(self, sample, labels_mask, lines_cols = None):
        # Extract the labels column
        cluster_data = sample.data[labels_mask]

        if lines_cols is not None:
            line_indices = np.unique(cluster_data[:, lines_cols])
            lines = sample.attrs["_lines"].lines
            cluster_lines = lines[line_indices.astype(int)]

        if self.extract_lines:
            return sample.attrs["_lines"].copy(data = cluster_lines)

        cluster = sample.copy(data = cluster_data)
        if lines_cols is not None:
            cluster.attrs["_lines"] = sample.attrs["_lines"].copy(
                data = cluster_lines,
            )

        return cluster


    def _empty_cluster(self, sample, lines_cols = None):
        if self.extract_lines:
            # Return empty LineData
            return sample.attrs["_lines"].copy(
                data = sample.attrs["_lines"][0:0],
            )

        cluster = sample.copy(data = sample[0:0])
        if lines_cols is not None:
            cluster.attrs["_lines"] = sample.attrs["_lines"].copy(
                data = sample.attrs["_lines"][0:0],
            )

        return cluster


    def fit_sample(self, sample: IterableSamples):
        if not isinstance(sample, IterableSamples):
            raise TypeError(textwrap.fill((
                "The input `sample` must be a subclass of `IterableSamples` "
                f"(e.g. PointData, LineData). Received type=`{type(sample)}`."
            )))

        # Extract the labels column
        col_idx = sample.columns.index("label")
        labels = sample.data[:, col_idx]

        # Check if there is a `._lines` attribute with `line_index` columns
        lines_cols = None
        if "_lines" in sample.attrs:
            lines_cols = [
                i for i, c in enumerate(sample.columns)
                if c.startswith("line_index")
            ]

            if len(lines_cols) == 0:
                warnings.warn((
                    "A `_lines` attribute was found, but no lines can "
                    "be extracted without columns `line_index<N>`."
                ), RuntimeWarning)

                lines_cols = None
                self.extract_lines = False

        elif self.extract_lines:
            raise ValueError(textwrap.fill((
                "If `extract_lines` is True, then the input `sample` must "
                "contain a `_lines` attribute."
            )))

        # If noise is requested, also include the noise cluster
        if self.noise:
            labels_unique = np.unique(labels)
        else:
            labels_unique = np.unique(labels[labels != -1])

        # For each unique label, create a new PointData / LineData cluster that
        # maintains / propagates all attributes (which needs a copy)
        clusters = [
            self._get_cluster(sample, labels == label, lines_cols)
            for label in labels_unique
        ]

        # If no valid cluster was found, return at least a single empty cluster
        if not len(clusters):
            clusters.append(self._empty_cluster(sample, lines_cols))

        # Remove the "label" column if needed
        if self.remove_labels and not self.extract_lines:
            for i in range(len(clusters)):
                clusters[i] = clusters[i].copy(
                    data = np.delete(clusters[i].data, col_idx, axis = 1),
                    columns = (clusters[i].columns[:col_idx] +
                               clusters[i].columns[col_idx + 1:]),
                )

        return clusters




def _wstd(x, w):
    '''Weighted standard deviation'''
    avg = np.average(x, weights = w)
    return np.sqrt(np.average((x - avg)**2, weights = w))




class Centroids(Filter):
    '''Compute the geometric centroids of a list of samples of points.

    Filter signature:

    ::

              PointData -> Centroids.fit_sample -> PointData
        list[PointData] -> Centroids.fit_sample -> PointData
          numpy.ndarray -> Centroids.fit_sample -> PointData

    This filter can be used right after ``pept.tracking.SplitLabels``, e.g.:

    >>> (SplitLabels() + Centroids()).fit(points)

    If `error = True`, append a measure of error on the computed centroid as
    the standard deviation in distances from centroid to all points. It is
    saved in an extra column "error".

    If `cluster_size = True`, append the number of points used for each
    centroid in an extra column "cluster_size" - unless `weight = True`, in
    which case it is the sum of weights.

    If `weight = True` and there is a column "weight" in the PointData, compute
    weighted centroids and standard deviations (if `error = True`) and the sum
    of weights (if `cluster_size = True`). The "weight" column is removed in
    the output centroid.

    '''


    def __init__(self, error = False, cluster_size = False, weight = True):
        self.error = bool(error)
        self.cluster_size = bool(cluster_size)
        self.weight = bool(weight)


    def _empty_centroid(self, points, weighted: bool):
        # Return an empty centroid with the correct number of columns
        ncols = points.shape[1]
        if self.error:
            ncols += 1
        if self.cluster_size:
            ncols += 1
        if weighted:
            ncols -= 1
        return np.empty((0, ncols))


    def _centroid(self, point_data, weighted: bool):
        # Extract the NumPy array of points
        points = point_data.points

        if len(points) == 0:
            return self._empty_centroid(points, weighted)

        if weighted:
            weightcol = point_data.columns.index("weight")
            weights = point_data.points[:, weightcol]

            # If all weights are zero, no cluster had been found
            if weights.sum() == 0.:
                return self._empty_centroid(points, weighted)

            c = np.average(points, weights = weights, axis = 0)
            c = np.delete(c, weightcol)
        else:
            c = points.mean(axis = 0)

        # If error is requested, compute std-dev of distances from centroid
        if self.error:
            d = np.linalg.norm(points[:, 1:4] - c[1:4], axis = 1)
            c = np.r_[c, _wstd(d, weights) if weighted else d.std()]

        # If cluster_size is requested, also append the number of points
        if self.cluster_size:
            c = np.r_[c, weights.sum() if weighted else len(points)]

        return c


    def fit_sample(self, points):
        # Type-checking inputs
        if isinstance(points, PointData):
            list_points = [points]
        elif isinstance(points, np.ndarray):
            list_points = [PointData(points)]
        else:
            list_points = [
                p if isinstance(p, PointData) else PointData(p)
                for p in list(points)
            ]

        if not len(list_points):
            raise ValueError("Must receive at least one PointData.")

        # If self.weight and there is a `weight` column, compute weighted
        # centroids
        weigh = (self.weight and "weight" in list_points[0].columns)

        # Compute centroid for each PointData and stack centroid arrays
        centroids = np.vstack([self._centroid(p, weigh) for p in list_points])
        attributes = list_points[0].extra_attrs()

        # If error or cluster_size are requested, append those columns
        columns = list_points[0].columns

        # Omit the `weight` column if weigh
        if weigh:
            weightcol = list_points[0].columns.index("weight")
            columns = columns[:weightcol] + columns[weightcol + 1:]

        if self.error:
            columns.append("error")

        if self.cluster_size:
            columns.append("cluster_size")

        return PointData(centroids, columns = columns, **attributes)




class LinesCentroids(Filter):
    '''Compute the minimum distance point of some ``pept.LineData`` while
    iteratively removing a fraction of the furthest lines.

    Filter signature:

    ::

        list[LineData] -> LinesCentroids.fit_sample -> PointData
              LineData -> LinesCentroids.fit_sample -> PointData
         numpy.ndarray -> LinesCentroids.fit_sample -> PointData

    The code below is adapted from the PEPT-EM algorithm developed by Antoine
    Renaud and Sam Manger.
    '''

    def __init__(self, remove = 0.1, iterations = 6):
        self.remove = float(remove)
        self.iterations = int(iterations)


    @staticmethod
    def centroid(lors):
        nx = np.newaxis

        m = np.identity(3)[nx, :, :] - lors[:, nx, 4:7] * lors[:, 4:7, nx]
        n = np.sum(m, axis = 0)
        v = np.sum(np.sum(m * lors[:, nx, 1:4], axis=-1), axis=0)

        return np.matmul(np.linalg.inv(n), v)


    @staticmethod
    def distance_matrix(x, lors):
        y = x[np.newaxis, :3] - lors[:, 1:4]
        return np.sum(y**2, axis=-1) - np.sum(y * lors[:, 4:7], axis=-1)**2


    def predict(self, lines):
        # Rewrite LoRs in the vectorial form y(x) = position + x * direction
        lors = lines[:, :7].copy(order = "C")

        lors[:, 4:7] = lors[:, 4:7] - lors[:, 1:4]
        lors[:, 4:7] /= np.linalg.norm(lors[:, 3:], axis = -1)[:, np.newaxis]

        # Begin with equal weights for all LoRs
        weights = np.ones(len(lors))
        x = LinesCentroids.centroid(lors)

        # Iteratively remove the furthest LoRs and recompute centroid
        for i in range(self.iterations):
            d2 = LinesCentroids.distance_matrix(x, lors)

            k = int(len(d2) * (1 - self.remove * (i + 1)))
            part = np.argpartition(d2, k)
            weights[part[k:]] = 0

            x = LinesCentroids.centroid(lors)

        # Add timestamp as the mean LoRs' time
        return np.hstack((lors[:, 0].mean(), x))


    def fit_sample(self, lines):
        # Type-checking inputs
        if isinstance(lines, LineData):
            list_lines = [lines]
        elif isinstance(lines, np.ndarray):
            list_lines = [LineData(lines)]
        else:
            list_lines = list(lines)

        centroids = [self.predict(lines.lines) for lines in list_lines]
        return PointData(np.vstack(centroids), **list_lines[0].extra_attrs())




class Condition(Filter):
    '''Select only data satisfying multiple conditions, given as a string, a
    function or list thereof; e.g. ``Condition("error < 15")`` selects all
    points whose "error" column value is smaller than 15.

    Filter signature:

    ::

        PointData -> Condition.fit_sample -> PointData
         LineData -> Condition.fit_sample -> LineData

    In the simplest case, a column name is specified, plus a comparison, e.g.
    ``Condition("error < 15, y > 100")``; multiple conditions may be
    concatenated using a comma.

    More complex conditions - where the column name is not the first operand -
    can be constructed using single quotes, e.g. using NumPy functions in
    ``Condition("np.isfinite('x')")`` to filter out NaNs and Infs. Quotes can
    be used to index columns too: ``Condition("'0' < 150")`` selects all rows
    whose first column is smaller than 150.

    Generally, you can use any function returning a boolean mask, either as a
    string of code ``Condition("np.isclose('x', 3)")`` or a user-defined
    function receiving a NumPy array ``Condition(lambda x: x[:, 0] < 10)``.

    Finally, multiple such conditions may be supplied separately:
    ``Condition(lambda x: x[:, -1] > 10, "'t' < 50")``.
    '''

    def __init__(self, *conditions):
        # Calls the conditions setter which does parsing
        self.conditions = conditions


    @property
    def conditions(self):
        return self._conditions


    @conditions.setter
    def conditions(self, conditions):
        if isinstance(conditions, str):
            self._conditions = Condition._parse_condition(conditions)
        elif callable(conditions):
            self._conditions = [conditions]
        else:
            cs = []
            for cond in conditions:
                cs.extend(Condition._parse_condition(cond))
            self._conditions = cs


    @staticmethod
    def _parse_condition(cond):
        if callable(cond):
            return [cond]

        conditions = str(cond).replace(" ", "").split(",")

        # Compile regex object to find quoted strings
        finder = re.compile(r"'\w+'")

        for i in range(len(conditions)):
            # Replace single-quoted column numbers / names
            if "'" in conditions[i]:
                conditions[i] = finder.sub(
                    Condition._replace_quoted,
                    conditions[i],
                )
                continue

            # If condition is a simple comparison, allow using non-quoted
            # column names
            op = None
            if "<" in conditions[i]:
                op = "<"
            elif ">" in conditions[i]:
                op = ">"
            elif "!" in conditions[i]:
                op = "!"
            elif "==" in conditions[i]:
                op = "=="

            if op is not None:
                cs = conditions[i].split(op)
                cs[0] = Condition._replace_term(cs[0])
                conditions[i] = op.join(cs)

            else:
                raise ValueError(textwrap.fill((
                    f"The input `conditions[i] = {conditions[i]}` did not "
                    "contain an operator or single-quoted terms."
                )))

        return conditions


    @staticmethod
    def _replace_term(term: str):
        return f"data[:, sample.columns.index('{term}')]"


    @staticmethod
    def _replace_quoted(term):
        # Remove single quotes
        if isinstance(term, re.Match):
            term = term.group()
        term = term.split("'")[1]

        try:
            index = int(term)
            return f"data[:, {index}]"
        except ValueError:
            return f"data[:, sample.columns.index('{term}')]"


    def fit_sample(self, sample: IterableSamples):
        if not isinstance(sample, IterableSamples):
            raise TypeError(textwrap.fill((
                "The input `sample` must be a subclass of `IterableSamples` "
                f"(e.g. PointData, LineData). Received type=`{type(sample)}`."
            )))

        data = sample.data

        for cond in self.conditions:
            if callable(cond):
                data = data[cond(data)]
            else:
                data = data[eval(cond, globals(), locals())]

        return sample.copy(data = data)




class SamplesCondition(Reducer):
    '''Select only *samples* satisfying multiple conditions, given as a string,
    a function or list thereof; e.g. ``Condition("sample_size > 30")`` selects
    all samples with a sample size larger than 30.

    Filter signature:

    ::

        PointData -> SamplesCondition.fit_sample -> PointData
         LineData -> SamplesCondition.fit_sample -> LineData

    This is different to a `Condition`, which selects individual points; for
    `SamplesCondition`, each sample will be passed through the conditions.

    Conditions can be defined as Python code using the following variables:

    - `sample` - this is the full PointData or LineData, e.g. only keep samples
      with more than 30 points with "len(sample.points) > 30".
    - `data` - this is the raw NumPy array of data wrapped by a PointData or
      LineData, e.g. only keep samples which have all X coordinates beyond 100
      with `SamplesCondition("np.all(data[:, 1] > 100)")`.
    - `sample_size` - this is a shorthand for the number of data points, e.g.
      only keep samples with more than 30 points with "sample_size > 30".

    Conditions can also be Python functions:

    >>> def high_velocity_filter(sample):
    >>>     return np.all(sample["v"] > 5)

    >>> from pept.tracking import SamplesCondition
    >>> filtered = SamplesCondition(high_velocity_filter).fit(point_data)

    '''

    def __init__(self, *conditions):
        # Calls the conditions setter which does parsing
        self.conditions = conditions


    @property
    def conditions(self):
        return self._conditions


    @conditions.setter
    def conditions(self, conditions):
        if isinstance(conditions, str):
            self._conditions = SamplesCondition._parse_condition(conditions)
        elif callable(conditions):
            self._conditions = [conditions]
        else:
            cs = []
            for cond in conditions:
                cs.extend(SamplesCondition._parse_condition(cond))
            self._conditions = cs


    @staticmethod
    def _parse_condition(cond):
        if callable(cond):
            return [cond]

        conditions = str(cond).replace(" ", "").split(",")

        # Compile regex object to find quoted strings
        finder = re.compile(r"'\w+'")

        for i in range(len(conditions)):
            # Replace single-quoted column numbers / names
            if "'" in conditions[i]:
                conditions[i] = finder.sub(
                    SamplesCondition._replace_quoted,
                    conditions[i],
                )
                continue

        return conditions


    @staticmethod
    def _replace_quoted(term):
        # Remove single quotes
        if isinstance(term, re.Match):
            term = term.group()
        term = term.split("'")[1]

        try:
            index = int(term)
            return f"data[:, {index}]"
        except ValueError:
            return f"data[:, sample.columns.index('{term}')]"


    def fit(self, samples):
        # Filtered samples
        collected = []

        for i, sample in enumerate(samples):
            # Defining variables to be accessed inside conditions
            if isinstance(sample, IterableSamples):
                data = sample.data
            else:
                data = sample

            sample_size = len(data)

            # Check all conditions are true
            keep = True
            for cond in self.conditions:
                if callable(cond):
                    keep = keep and cond(sample)
                else:
                    keep = keep and eval(cond, globals(), locals())

            if keep:
                collected.append(sample)
            elif len(collected) == 0 and i == len(samples) - 1:
                # If no sample was retained, save an empty sample
                collected.append(sample[0:0])

        return collected




class Remove(Filter):
    '''Remove columns (either column names or indices) from `pept.LineData` or
    `pept.PointData`.

    Filter signature:

    ::

         pept.LineData -> Remove.fit_sample -> pept.LineData
        pept.PointData -> Remove.fit_sample -> pept.PointData

    Examples
    --------
    To remove a single column named "line_index":

    >>> import pept
    >>> from pept.tracking import *
    >>> points = pept.PointData(...)    # Some dummy data

    >>> rem = Remove("line_index")
    >>> points_without = rem.fit_sample(points)

    Remove all columns starting with "line_index" using a glob operator (*):

    >>> points_without = Remove("line_index*").fit_sample(points)

    Remove the first column based on its index:

    >>> points_without = Remove(0).fit_sample(points)

    Finally, multiple removals may be chained into a list:

    >>> points_without = Remove(["line_index*", -1]).fit_sample(points)

    '''

    def __init__(self, *columns):
        self._indices = []
        self._filters = []

        # Calls the `columns` setter which does parsing
        self.columns = columns


    @property
    def columns(self):
        return self._columns


    @columns.setter
    def columns(self, columns):
        self._columns = [Remove._parse(col) for col in columns]

        # Split the removers into regex strings and column indices
        for c in self._columns:
            if isinstance(c, str):
                self._filters.append(c)
            else:
                self._indices.append(c)


    @staticmethod
    def _parse(col):
        if isinstance(col, str):
            return col.replace("*", r"\w*")
        elif isinstance(col, Number):
            return int(col)
        else:
            raise ValueError(textwrap.fill((
                "Each input argument in `columns` must be a string or an "
                f"integer. One of them was `type(col) = {type(col)}`."
            )))


    def fit_sample(self, sample: IterableSamples):
        if not isinstance(sample, IterableSamples):
            raise TypeError(textwrap.fill((
                "The input `sample` must be a subclass of `IterableSamples` "
                f"(e.g. PointData, LineData). Received type=`{type(sample)}`."
            )))

        # Extract the relevant `sample` attributes
        columns = sample.columns
        ncols = len(columns)

        # The regex filters to use and column numbers to remove
        filters = self._filters
        indices = self._indices

        # Column indices to remove and remaining column names
        removed = set()
        columns_filtered = []

        for i, c in enumerate(columns):
            # Also handle negative indices
            if any((re.fullmatch(r, c) for r in filters)) or \
                    any((i == ind or i == ind + ncols for ind in indices)):
                removed.add(i)
            else:
                columns_filtered.append(c)

        indices_filtered = [i for i in range(len(columns)) if i not in removed]
        data = sample.data[:, indices_filtered]

        return sample.copy(data = data, columns = columns_filtered)




class SplitAll(Reducer):
    '''Stack all samples and split them into a list according to a named /
    numeric column index.

    Reducer signature:

    ::

                   LineData -> SplitAll.fit -> list[LineData]
             list[LineData] -> SplitAll.fit -> list[LineData]

                  PointData -> SplitAll.fit -> list[PointData]
            list[PointData] -> SplitAll.fit -> list[PointData]

              numpy.ndarray -> SplitAll.fit -> list[numpy.ndarray]
        list[numpy.ndarray] -> SplitAll.fit -> list[numpy.ndarray]

    If using a LineData / PointData, you can use a columns name as a string,
    e.g. ``SplitAll("label")`` or a number ``SplitAll(4)``. If using a NumPy
    array, only numeric indices are accepted.
    '''

    def __init__(self, column):
        try:
            self.column_index = int(column)
            self.column_name = None
        except ValueError:
            self.column_name = str(column)
            self.column_index = None


    def fit(self, samples):
        if not hasattr(samples, "__iter__"):
            raise TypeError(textwrap.fill((
                "The input `samples` must be an iterable (e.g. list, tuple, "
                f"PointData, LineData). Received type=`{type(samples)}`."
            )))

        # Reduce / stack list of samples onto a single IterableSamples / array
        samples = Stack().fit(samples)

        if isinstance(samples, np.ndarray):
            return self._split_numpy(samples)
        elif isinstance(samples, IterableSamples):
            return self._split_iterable_samples(samples)
        else:
            raise TypeError(textwrap.fill((
                "The input samples must be NumPy arrays, PointData / LineData "
                f"or lists thereof. Received `type(samples) = {type(samples)}`"
            )))


    def _split_numpy(self, samples):
        if self.column_index is None:
            raise TypeError(textwrap.fill((
                "If the samples are NumPy arrays, you must use a numeric "
                f"column index; used a named column: `{self.column_name}`."
            )))

        col_data = samples[:, self.column_index]
        labels = np.unique(col_data)

        # If no labels exist, return a list with an empty sample
        if not len(labels):
            return [samples[0:0]]

        return [samples[col_data == label] for label in labels]


    def _split_iterable_samples(self, samples):
        if self.column_index is not None:
            col_data = samples.data[:, self.column_index]
        else:
            col_data = samples.data[:, samples.columns.index(self.column_name)]

        labels = np.unique(col_data)

        # If no labels exist, return a list with an empty sample
        if not len(labels):
            return [samples[0:0]]

        return [
            samples.copy(data = samples.data[col_data == label])
            for label in labels
        ]


# Use standard names, Leonard...
GroupBy = SplitAll




class Swap(Filter):
    '''Swap two columns in a LineData or PointData.

    Filter signature:

    ::

         LineData -> Swap.fit_sample -> LineData
        PointData -> Swap.fit_sample -> PointData

    For example, swap the Y and Z axes: ``Swap("y, z").fit_sample(points)``.
    Add multiple swaps as separate arguments: ``Swap("y, z", "label, x")``.

    You can also swap columns at numerical indices by single-quoting them:
    ``Swap("'0', '1'")``.

    *New in pept-0.4.3*
    '''

    def __init__(self, *swaps, inplace = True):
        # Calls swaps.setter which does type-checking
        self.swaps = swaps
        self.inplace = bool(inplace)


    @property
    def swaps(self):
        return self._swaps


    @swaps.setter
    def swaps(self, swaps):
        commands = []

        # Compile regex object to find quoted strings
        finder = re.compile(r"'\w+'")

        for i in range(len(swaps)):
            s = swaps[i].replace(" ", "").split(",")
            if len(s) != 2:
                raise ValueError(textwrap.fill((
                    f"The input `swaps[{i}] = {swaps[i]}` does not have a "
                    'single comma. It must be formatted as "col1, col2".'
                )))

            # Replace single-quoted column numbers / names
            for j in range(2):
                if "'" in s[j]:
                    s[j] = finder.sub(Swap._replace_quoted, s[j])
                else:
                    s[j] = Swap._replace_term(s[j])

            commands.append(
                f"aux = {s[0]}.copy() ; {s[0]} = {s[1]} ; {s[1]} = aux;\n"
            )

        self._swaps = tuple(commands)


    @staticmethod
    def _replace_term(term: str):
        return f"data[:, sample.columns.index('{term}')]"


    @staticmethod
    def _replace_quoted(term):
        # Remove single quotes
        if isinstance(term, re.Match):
            term = term.group()
        term = term.split("'")[1]

        try:
            index = int(term)
            return f"data[:, {index}]"
        except ValueError:
            return f"data[:, sample.columns.index('{term}')]"


    def fit_sample(self, sample: IterableSamples):
        if not isinstance(sample, IterableSamples):
            raise TypeError(textwrap.fill((
                "The input `sample` must be a subclass of `IterableSamples` "
                f"(e.g. PointData, LineData). Received type=`{type(sample)}`."
            )))

        if not self.inplace:
            sample = sample.copy()

        data = sample.data
        for s in self.swaps:
            exec(s, locals(), globals())

        return sample




class OptimizeWindow(Reducer):
    '''Automatically determine optimum adaptive time window to have an ideal
    number of elements per sample.

    Reducer signature:

    ::

               LineData -> OptimizeWindow.fit -> LineData
         list[LineData] -> OptimizeWindow.fit -> LineData
              PointData -> OptimizeWindow.fit -> PointData
        list[PointData] -> OptimizeWindow.fit -> PointData
          numpy.ndarray -> OptimizeWindow.fit -> PointData

    The adaptive time window approach combines the advantages of fixed sample
    sizes and time windowing:

    - Time windows are robust to tracers moving in and out of the field of
      view, as they simply ignore the time slices where almost no LoRs are
      recorded.
    - Fixed sample sizes effectively adapt their spatio-temporal resolution,
      allowing for higher accuracy when tracers are passing through more
      active scanner regions.

    All samples with more than `ideal_elems` are shortened, such that time
    windows are shrinked when the tracer activity permits. There exists an
    ideal time window such that most samples will have roughly `ideal_elems`,
    with a few higher activity ones that are shortened; ``OptimizeWindow``
    finds this ideal time window for ``pept.AdaptiveWindow``.

    *New in pept-0.5.1*

    Examples
    --------
    Find an adaptive time window that is ideal for about 200 LoRs per sample:

    >>> import pept
    >>> import pept.tracking as pt
    >>> lors = pept.LineData(...)
    >>> lors = pt.OptimizeWindow(ideal_elems = 200).fit(lors)

    `OptimizeWindow` can be used at the start of a pipeline; an optional
    `overlap` parameter can be used to define an overlap as a ratio to the
    ideal time window found. For example, if the ideal time window found is
    100 ms, an overlap of 0.5 will result in an overlapping time interval of
    50 ms:

    >>> pipeline = pept.Pipeline([
    >>>     pt.OptimizeWindow(200, overlap = 0.5),
    >>>     pt.BirminghamMethod(0.5),
    >>>     pt.Stack(),
    >>> ])
    '''

    def __init__(self, ideal_elems, overlap = 0., low = 0.3, high = 3):
        self.ideal_elems = int(ideal_elems)

        overlap = float(overlap)
        if overlap >= 1 or overlap < 0:
            raise ValueError((
                "\n[ERROR]: If `overlap` is defined, it must be the ratio "
                "relative to the ideal time window that will be found, "
                f"and hence between [0, 1). Received {overlap}."
            ))
        self.overlap = overlap

        self.low = float(low)
        self.high = float(high)


    def fit(self, data):

        # Stack all data into LineData or PointData (default)
        data = Stack().fit(data)
        if not isinstance(data, PEPTObject):
            data = PointData(data)

        self.data = data

        # Compute bounds
        times = data.data[:, 0]
        dt = times[1:] - times[:-1]
        estimate = self.ideal_elems * np.nanmedian(dt)

        # In extreme cases, median(dt) is 0
        if estimate == 0.:
            for q in [0.8, 0.9, 0.95, 0.99, 0.999]:
                estimate = self.ideal_elems * np.nanquantile(dt, q)
                if estimate != 0.:
                    break
            else:
                raise ValueError("Ideal time window could not be estimated.")

        # Find time window that yields median LoR counts per sample as close to
        # the `ideal_elems`
        res = minimize_scalar(
            self.evaluate,
            bracket = [0.5 * estimate, estimate],
        )

        # Set sample size and overlap (if requested) to ideal values found
        data.sample_size = AdaptiveWindow(res.x, self.ideal_elems)
        data.overlap = AdaptiveWindow(res.x * self.overlap)

        return data


    def evaluate(self, window):
        # Window cannot be negative; return maximum error
        if window < 0:
            return np.inf

        # Set adaptive window to compute samples indices
        self.data.sample_size = AdaptiveWindow(window)

        # Compute number of counts per sample
        si = self.data.samples_indices
        counts = np.array(si[:, 1] - si[:, 0], dtype = float)

        # Ignore samples where the tracer is outside the system (very low
        # counts) or is extremely active / static
        counts[counts < self.low * self.ideal_elems] = np.nan
        counts[counts > self.high * self.ideal_elems] = np.nan

        return (np.nanmedian(counts) - self.ideal_elems) ** 2




class Debug(Reducer):
    '''Print types and statistics about the objects being processed in a
    ``pept.Pipeline``.

    Reducer signature:

    ::

              PointData -> Debug.fit -> PointData
               LineData -> Debug.fit -> LineData
        list[PointData] -> Debug.fit -> list[PointData]
         list[LineData] -> Debug.fit -> list[LineData]
             np.ndarray -> Debug.fit -> np.ndarray
                    Any -> Debug.fit -> Any

    This is a reducer, so it will collect all samples processed up to the
    point of use, print them, and return them unchanged.

    *New in pept-0.5.1*

    Examples
    --------
    A ``Debug`` is normally added in a ``Pipeline``:

    >>> import pept
    >>> import pept.tracking as pt
    >>>
    >>> pept.Pipeline([
    >>>     # First pass of clustering
    >>>     pt.Cutpoints(max_distance = 0.2),
    >>>     pt.HDBSCAN(true_fraction = 0.15),
    >>>     pt.SplitLabels() + pt.Centroids(cluster_size = True, error = True),
    >>>
    >>>     pt.Debug(),
    >>>     pt.Stack(),
    >>> ])
    '''

    def __init__(self, verbose = 5, max_samples = 10):
        self.verbose = int(verbose)
        self.max_samples = int(max_samples)


    def _print_stats_pla(self, samples):
        # Printing statistics for PointData, LineData, np.ndarray
        if isinstance(samples, np.ndarray):
            data = samples
            columns = None
        else:
            data = samples.data
            columns = samples.columns

        if self.verbose >= 1:
            print(samples)

        # Create summary statistics using pandas; add new summary columns
        if self.verbose >= 2:
            df = pd.DataFrame(data, columns = columns)

            desc = df.describe(include = "all")
            desc.loc['dtype'] = df.dtypes
            desc.loc['NaN'] = df.isnull().sum()

            print(desc)


    def fit(self, samples):
        # Lines for printing
        over = "=" * 80
        under = "-" * 80
        print("\n" + over + "\nDebug Start\n" + under)

        # Special-cased types
        pla = (PointData, LineData, np.ndarray)

        if isinstance(samples, list):
            if len(samples) == 0:
                print("No samples given, received empty list!")
            else:
                # Print unique types in list
                unique_types = set(type(s) for s in samples)
                print(f"Processing {len(samples)} samples of:")
                for ut in unique_types:
                    print(" ", ut)

                # If list contains PointData, LineData or simple NumPy arrays,
                # stack them and print statistics via pandas
                if len(unique_types) == 1 and isinstance(samples[0], pla):
                    print("Stacking data to print statistics...\n")

                    # Reduce / stack list of samples onto a single object
                    stacked = Stack().fit(samples)
                    self._print_stats_pla(stacked)

                # Unknown types in list
                else:
                    print(f"Printing the first {self.max_samples} samples")
                    for i, s in samples:
                        print(f"\nSample {i}:\n{under}")
                        print(s)

        elif isinstance(samples, pla):
            print(f"Processing a single {type(samples)}:")
            self._print_stats_pla(samples)

        else:
            print(f"Processing {type(samples)}:")
            print(samples)

        print("\n" + under + "\nDebug End\n" + over)
        return samples



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : utilities.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 08.08.2021


import textwrap


def check_homogeneous_types(iterable):
    if len(iterable) == 0:
        return

    base_type = type(iterable[0])
    for i, e in enumerate(iterable):
        if not isinstance(e, base_type):
            raise TypeError(textwrap.fill((
                "The input iterable must have homogeneous types. The first "
                f"element was of type `{base_type}`, but the element at index "
                f"{i} was of type `{type(e)}`."
            )))


def check_iterable(target_type, **kwargs):
    '''Check that an iterable has elements of type `target_type`.

    Performance caveat: only checks the first element. If the iterable is empty
    it passes.

    Raises
    ------
    TypeError
        If the first keyword argument value is not an iterable or its first
        element is not an object of type `target_type`.

    Examples
    --------
    >>> check_iterable(PointData, samples = [PointData(...), PointData(...)])
    '''
    # Extract the first keyword argument name and value
    for obj, val in kwargs.items():
        break

    if not hasattr(val, "__iter__"):
        raise TypeError(textwrap.fill((
            f"The input `{obj}` must be an iterable (list, tuple, PointData, "
            f"LineData, etc.). Received type=`{type(val)}`."
        )))

    if len(val) and not isinstance(val[0], target_type):
        raise TypeError(textwrap.fill((
            f"The input `{obj}` must be an iterable containing elements of "
            f"type `{target_type}`. The first element in `{obj}` was of type "
            f"`{type(val[0])}`."
        )))


def memoryview_safe(x):
    """Make array safe to run in a Cython memoryview-based kernel. These
    kernels typically break down with the error ``ValueError: buffer source
    array is read-only`` when running in dask distributed or joblib.

    Taken from `https://github.com/dask/distributed/issues/1978`.
    """
    if not x.flags.writeable:
        if not x.flags.owndata:
            x = x.copy(order='C')
        x.setflags(write=True)
    return x



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


# File   : __init__.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 22.06.2020


'''The PEPT-oriented post-processing suite, including occupancy grid,
vector velocity fields, etc.

This module contains fast, robust functions that operate on PEPT-like data
and integrate with the `pept` library's base classes.

'''


from    .grids  import  DynamicProbability2D, ResidenceDistribution2D
from    .grids  import  DynamicProbability3D, ResidenceDistribution3D
from    .grids  import  VectorField2D, VectorField3D
from    .grids  import  VectorGrid2D, VectorGrid3D

from    .mixing import  LaceyColors, LaceyColorsLinear
from    .mixing import  RelativeDeviations, RelativeDeviationsLinear
from    .mixing import  AutoCorrelation
from    .mixing import  SpatialProjections


__license__ = "GNU v3.0"
__maintainer__ = "Andrei Leonard Nicusan"
__email__ = "a.l.nicusan@bham.ac.uk"
__status__ = "Beta"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : __init__.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 31.01.2020


'''PEPT base classes.
'''


from    .line_data          import  LineData
from    .point_data         import  PointData

from    konigcell           import  Pixels
from    konigcell           import  Voxels

from    .iterable_samples   import  IterableSamples
from    .iterable_samples   import  TimeWindow
from    .iterable_samples   import  AdaptiveWindow
from    .iterable_samples   import  AsyncIterableSamples
from    .iterable_samples   import  PEPTObject

from    .pipelines          import  Transformer
from    .pipelines          import  Filter
from    .pipelines          import  LineDataFilter
from    .pipelines          import  PointDataFilter
from    .pipelines          import  VoxelsFilter
from    .pipelines          import  Reducer
from    .pipelines          import  Pipeline

from    .utilities          import  check_iterable


# Execute code here to add dynamic methods
from    .                   import  pixels
from    .                   import  voxels



__license__ = "GNU v3.0"
__maintainer__ = "Andrei Leonard Nicusan"
__email__ = "a.l.nicusan@bham.ac.uk"
__status__ = "Beta"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : base.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 29.09.2021


import  textwrap

import  numpy           as      np
from    scipy.optimize  import  minimize

from    pept            import  LineData, PointData
from    pept.base       import  LineDataFilter, Filter

from    ..peptml        import  get_cutoffs
from    .cutpoints_tof  import  find_cutpoints_tof




class TimeOfFlight(LineDataFilter):
    '''Compute the positron annihilation locations of each LoR as given by the
    Time Of Flight (ToF) data of the two LoR timestamps.

    Filter signature:
    ::

        LineData -> TimeOfFlight.fit_sample -> LineData  (points = False)
        LineData -> TimeOfFlight.fit_sample -> PointData (points = True)

    Importantly, the input LineData must have at least 8 columns, formatted as
    [t1, x1, y1, z1, x2, y2, z2, t2] - notice the different timestamps of the
    two LoR ends.

    If `points = False` (default), the computed ToF points are saved as an
    extra attribute "tof" in the input LineData; otherwise they are returned
    directly.

    The `temporal_resolution` should be set to the FWHM of the temporal
    resolution in the LoR timestamps, in self-consistent units of measure (e.g.
    m / s or mm / ms, but not mm / s). If it is set, the "temporal_resolution"
    and "spatial_resolution" extra attributes are set on the ToF points.

    *New in pept-0.4.2*

    Examples
    --------
    Generate 10 random LoRs between (-100, 100) mm and ms with 8 columns for
    the ToF format.

    >>> import numpy as np
    >>> import pept

    >>> rng = np.random.default_rng(0)
    >>> lors = pept.LineData(
    >>>     rng.uniform(-100, 100, (10, 8)),
    >>>     columns = ["t1", "x1", "y1", "z1", "x2", "y2", "z2", "t2"],
    >>> )
    >>> lors
    pept.LineData (samples: 1)
    --------------------------
    sample_size = 10
    overlap = 0
    lines =
      (rows: 10, columns: 8)
      [[ 57.4196615  -52.1261114  ...  -9.93212667  59.26485406]
       [-53.8715582  -89.59573979 ... -40.26077344  34.39897559]
       ...
       [ 51.59020047   2.55174465 ... -31.13800424 -13.94025361]
       [ 93.21241616  12.44636845 ... -75.08905883 -42.3338486 ]]
    columns = ['t1', 'x1', 'y1', 'z1', 'x2', 'y2', 'z2', 't2']
    attrs = {}

    Compute Time of Flight annihilation locations from the two timestamps
    in the data above. Assume a temporal resolution of 100 ps - be careful to
    use self-consistent units; in this case we are using mm and ms:

    >>> from pept.tracking import *

    >>> temporal_resolution = 100e-12 * 1000    # ms
    >>> lors_tof = TimeOfFlight(temporal_resolution).fit_sample(lors)
    >>> lors_tof
    pept.LineData (samples: 1)
    --------------------------
    sample_size = 10
    overlap = 0
    lines =
      (rows: 10, columns: 8)
      [[ 57.4196615  -52.1261114  ...  -9.93212667  59.26485406]
       [-53.8715582  -89.59573979 ... -40.26077344  34.39897559]
       ...
       [ 51.59020047   2.55174465 ... -31.13800424 -13.94025361]
       [ 93.21241616  12.44636845 ... -75.08905883 -42.3338486 ]]
    columns = ['t1', 'x1', 'y1', 'z1', 'x2', 'y2', 'z2', 't2']
    attrs = {
      'tof': pept.PointData (samples: 1)
    ---------------------------
    sample_...
    }

    >>> lors_tof.attrs["tof"]
    pept.PointData (samples: 1)
    ---------------------------
    sample_size = 10
    overlap = 0
    points =
      (rows: 10, columns: 4)
      [[ 5.64970655e+01 -3.22092074e+07  2.41767704e+08 -1.30428351e+08]
       [-9.80068250e+01 -2.48775932e+09 -1.12904720e+10 -6.43480969e+09]
       ...
       [ 1.88249731e+01  3.34819602e+09 -8.78848458e+09  2.83529405e+09]
       [ 2.54392837e+01  1.90343279e+10 -1.92717662e+09 -6.84078611e+09]]
    columns = ['t', 'x', 'y', 'z']
    attrs = {
      'temporal_resolution': 1.0000000000000001e-07
      'spatial_resolution': 29.9792458
    }

    Alternatively, you can extract only the ToF points directly:

    >>> tof = TimeOfFlight(temporal_resolution, points = True).fit_sample(lors)
    >>> tof
    pept.PointData (samples: 1)
    ---------------------------
    sample_size = 10
    overlap = 0
    points =
      (rows: 10, columns: 4)
      [[ 5.64970655e+01 -3.22092074e+07  2.41767704e+08 -1.30428351e+08]
       [-9.80068250e+01 -2.48775932e+09 -1.12904720e+10 -6.43480969e+09]
       ...
       [ 1.88249731e+01  3.34819602e+09 -8.78848458e+09  2.83529405e+09]
       [ 2.54392837e+01  1.90343279e+10 -1.92717662e+09 -6.84078611e+09]]
    columns = ['t', 'x', 'y', 'z']
    attrs = {
      'temporal_resolution': 1.0000000000000001e-07
      'spatial_resolution': 29.9792458
    }

    '''

    def __init__(self, temporal_resolution = None, points = False):
        if temporal_resolution is None:
            self.temporal_resolution = None
        else:
            self.temporal_resolution = float(temporal_resolution)

        self.points = bool(points)


    def fit_sample(self, sample: LineData):
        if not isinstance(sample, LineData):
            sample = LineData(sample)

        if sample.lines.shape[1] < 8:
            raise ValueError(textwrap.fill((
                "The input `sample` of LineData must have at least 8 columns, "
                "formatted as [t1, x1, y1, z1, x2, y2, z2, t2]. Received "
                f"sample with `sample.lines.shape = {sample.lines.shape}`."
            )))

        # The two points defining the LoR
        t1 = sample.lines[:, 0]
        p1 = sample.lines[:, 1:4]

        p2 = sample.lines[:, 4:7]
        t2 = sample.lines[:, 7]

        # Speed of light (mm / ms)
        c = 299792458.

        # The ratio (P1 - tofpoint) / (P1 - P2) for all rows
        distance_ratio = (
            0.5 - 0.5 / np.linalg.norm(p2 - p1, axis = 1) * c * (t2 - t1)
        )

        # [:, np.newaxis] = transform row vector to column vector (i.e. 2D
        # array with one column)
        tof_locations = p1 + (p2 - p1) * distance_ratio[:, np.newaxis]
        tof_time = t1 - np.linalg.norm(tof_locations - p1, axis = 1) / c

        # Encapsulate ToF points in a PointData. Save resolution as attrs
        tof_pts = PointData(np.c_[tof_time, tof_locations])

        if self.temporal_resolution is not None:
            tof_pts.attrs["temporal_resolution"] = self.temporal_resolution
            tof_pts.attrs["spatial_resolution"] = self.temporal_resolution * c

        if self.points:
            return tof_pts

        sample.attrs["tof"] = tof_pts
        return sample




class CutpointsToF(LineDataFilter):
    '''Compute cutpoints from all pairs of lines whose Time Of Flight-predicted
    locations are closer than `max_distance`.

    Filter signature:
    ::

        LineData -> CutpointsToF.fit_sample -> PointData

    If the ``TimeOfFlight`` filter was used and a temporal resolution was
    specified (as a FWHM), then `max_distance` is automatically inferred as
    the minimum between 2 * "spatial_resolution" and the dimension-wise
    standard deviation of the input points.

    The `cutoffs` parameter may be set as [xmin, xmax, ymin, ymax, zmin, zmax]
    for a minimum bounding box outside of which cutpoints are discarded.
    Otherwise it is automatically set to the minimum bounding box containing
    all input LoRs.

    If `append_indices = True`, two extra columns are appended to the result as
    "line_index1" and "line_index2" containing the indices of the LoRs that
    produced each cutpoint; an extra attribute "_lines" is also set to the
    input `LineData`.

    If `cutpoints_only = False` (default), the Time Of Flight-predicted
    positron annihilation locations are also appended to the returned points.

    *New in pept-0.4.2*

    See Also
    --------
    pept.LineData : Encapsulate LoRs for ease of iteration and plotting.
    pept.tracking.HDBSCAN : Efficient, HDBSCAN-based clustering of (cut)points.
    pept.read_csv : Fast CSV file reading into numpy arrays.

    Examples
    --------
    Make sure to use the ``TimeOfFlight`` filter to compute to ToF annihilation
    locations; if you specify a temporal resolution, the `max_distance`
    parameter is automatically computed:

    >>> import pept
    >>> from pept.tracking import *

    >>> line_data = pept.LineData(example_tof_data)
    >>> line_data_tof = TimeOfFlight(100e-9).fit_sample(line_data)
    >>> cutpoints_tof = CutpointsToF().fit_sample(line_data_tof)

    Alternatively, set `max_distance` yourself:

    >>> line_data = pept.LineData(example_tof_data)
    >>> line_data_tof = TimeOfFlight().fit_sample(line_data)
    >>> cutpoints_tof = CutpointsToF(5.0).fit_sample(line_data_tof)
    '''

    def __init__(
        self,
        max_distance = None,
        cutoffs = None,
        append_indices = False,
        cutpoints_only = False,
    ):
        # Setting class attributes. The ones below call setters which do type
        # checking
        self.cutoffs = cutoffs
        self.append_indices = append_indices
        self.max_distance = max_distance

        self.cutpoints_only = bool(cutpoints_only)


    @property
    def max_distance(self):
        return self._max_distance


    @max_distance.setter
    def max_distance(self, max_distance):
        if max_distance is None:
            self._max_distance = None
        else:
            self._max_distance = float(max_distance)


    @property
    def cutoffs(self):
        return self._cutoffs


    @cutoffs.setter
    def cutoffs(self, cutoffs):
        if cutoffs is not None:
            cutoffs = np.asarray(cutoffs, order = 'C', dtype = float)
            if cutoffs.ndim != 1 or len(cutoffs) != 6:
                raise ValueError((
                    "\n[ERROR]: cutoffs should be a one-dimensional array "
                    "with values [min_x, max_x, min_y, max_y, min_z, max_z]. "
                    f"Received {cutoffs}.\n"
                ))

            self._cutoffs = cutoffs
        else:
            self._cutoffs = None


    @property
    def append_indices(self):
        return self._append_indices


    @append_indices.setter
    def append_indices(self, append_indices):
        self._append_indices = bool(append_indices)


    def fit_sample(self, sample_lines: LineData):
        if not isinstance(sample_lines, LineData):
            sample_lines = LineData(sample_lines)

        # Ensure ToF data is present and extract it
        if "tof" not in sample_lines.attrs:
            raise AttributeError(textwrap.fill((
                "The input `sample_lines` must have an attribute 'tof' "
                "containing the Time of Flight predicted annihilation point "
                "for each LoR. This is normally set by "
                "`pept.tracking.TimeOfFlight`."
            )))

        tofpoints = sample_lines.attrs["tof"]
        tofattrs = tofpoints.attrs

        # If max_distance was not defined, use FWHM spatial_resolution
        if self.max_distance is not None:
            max_distance = self.max_distance
        elif "spatial_resolution" in tofpoints.attrs:
            max_distance = min(
                tofpoints.attrs["spatial_resolution"] * 2,
                tofpoints.points[:, 1:4].std(axis = 0).max(),
            )
        else:
            max_distance = tofpoints.points[:, 1:4].std(axis = 0).max()

        # If cutoffs were not defined, automatically compute them
        if self.cutoffs is not None:
            cutoffs = self.cutoffs
        else:
            cutoffs = get_cutoffs(sample_lines.lines)

        # Only compute cutpoints if there are at least 2 LoRs
        if len(sample_lines.lines) >= 2:
            sample_cutpoints = find_cutpoints_tof(
                sample_lines.lines,
                tofpoints.points,
                max_distance,
                cutoffs,
                append_indices = self.append_indices,
            )
        else:
            sample_cutpoints = np.empty((0, 6 if self.append_indices else 4))

        # Column names
        columns = ["t", "x", "y", "z"]
        if self.append_indices:
            columns += ["line_index1", "line_index2"]

        # Also include tofpoints if not self.cutpoints_only
        if self.cutpoints_only:
            points = PointData(sample_cutpoints, columns = columns, **tofattrs)
        else:
            tofs = tofpoints.points
            if self.append_indices:
                tofs = np.c_[tofs, np.arange(len(tofs)), np.arange(len(tofs))]

            points = PointData(
                np.vstack((sample_cutpoints, tofs)),
                columns = columns,
                **tofattrs,
            )

        # Add optional metadata to the points; because they have an underscore,
        # they won't be propagated when new objects are constructed
        points.attrs["_max_distance"] = max_distance
        points.attrs["_cutoffs"] = cutoffs

        if self.append_indices:
            points.attrs["_lines"] = sample_lines

        return points




def _gaussian_probs(x, points, sigma):
    '''Probabilities from x to Gaussian-distributed points.'''
    dists = np.sum((x[:3] - points)**2, axis = 1)
    return np.exp(-0.5 * dists / sigma**2)




class GaussianDensity(Filter):
    '''Append weights according to the Gaussian distribution that best fits
    the samples of points.

    Filter signature:
    ::

              PointData -> GaussianDensity.fit_sample -> PointData
          numpy.ndarray -> GaussianDensity.fit_sample -> PointData
        list[PointData] -> GaussianDensity.fit_sample -> list[PointData]

    This is treated as an optimisation problem: find the 3D location that
    maximises the sum of Probability Distributions (PDF) centered at each
    point.

    ::

        Given N points p_1, p_2, ..., p_N:

                  N
        maximise sum( exp( -0.5 * |x - p_i|^2 / sigma^2 ) )
           x      i

    Each point is then assigned a weight corresponding to its PDF - i.e. the
    exponential part - saved in the `weight` column.

    Sigma controls the standard deviation of the Gaussian distribution centred
    at each point; this corresponds to the relative uncertainty in each point's
    location. For ``TimeOfFlight`` data, leave `sigma = None` and it will be
    computed from the "spatial_resolution" attribute.

    You can use ``Centroids`` afterwards to compute the weighted centroid, i.e.
    where the tracer is. For multiple particle tracking (or just more
    robustness to noise) you can use `HDBSCAN + SplitLabels` beforehand.

    *New in pept-0.4.2*
    '''

    def __init__(self, sigma = None):
        self.sigma = None if sigma is None else float(sigma)


    def _add_weights(self, point_data, sigma):
        # Extract the inner NumPy array of points
        points = point_data.points

        # If no points were given, still return an empty array with the correct
        # number of columns
        if len(points) == 0:
            return np.empty((0, points.shape[1] + 1))

        # Find the 3D point that maximises the sum of PDFs
        xyz = np.array(points[:, 1:4])

        res = minimize(
            lambda x: -np.sum(_gaussian_probs(x, xyz, sigma)),
            np.median(xyz, axis = 0),
            method = "SLSQP",
        )

        # The PDF of each point acts as a weight
        weights = _gaussian_probs(res.x, xyz, sigma)
        return np.c_[points, weights]


    def fit_sample(self, points):
        # Type-checking inputs
        return_list = False
        if isinstance(points, PointData):
            list_points = [points]
        elif isinstance(points, np.ndarray):
            list_points = [PointData(points)]
        else:
            return_list = True
            list_points = [
                p if isinstance(p, PointData) else PointData(p)
                for p in list(points)
            ]

        if not len(list_points):
            raise ValueError("Must receive at least one PointData.")

        # Ensure either self.sigma is set or points.attrs["spatial_resolution"]
        # exists
        if self.sigma is None and any(
            ("spatial_resolution" not in lp.attrs for lp in list_points)
        ):
            raise AttributeError(textwrap.fill((
                "If `sigma` is not set, `points.attrs['spatial_resolution']` "
                "is used. However, it was not found (it is normally set by "
                "`pept.tracking.TimeOfFlight` if `temporal_resolution` is "
                "given)."
            )))
        else:
            fwhms = 2 * np.sqrt(2 * np.log(2))
            self.sigma = list_points[0].attrs["spatial_resolution"] / fwhms

        # Add "weight" column to each PointData
        list_weighted = [self._add_weights(p, self.sigma) for p in list_points]
        columns = list_points[0].columns + ["weight"]

        if return_list:
            return [
                points.copy(data = weighted, columns = columns)
                for points, weighted in zip(list_points, list_weighted)
            ]

        return PointData(
            np.vstack(list_weighted),
            columns = columns,
            **list_points[0].attrs,
        )
